"""Generate specific tensor network states and operators.
"""
from numbers import Integral

import numpy as np

from ..core import make_immutable, ikron
from ..gen.operators import spin_operator, eye, _gen_mbl_random_factors
from ..gen.rand import randn, choice, random_seed_fn, rand_phase
from .tensor_core import Tensor
from .array_ops import asarray, sensibly_scale
from .tensor_1d import MatrixProductState, MatrixProductOperator
from .tensor_tebd import NNI


@random_seed_fn
def rand_tensor(shape, inds, tags=None, dtype=float, left_inds=None):
    """Generate a random tensor with specified shape and inds.

    Parameters
    ----------
    shape : sequence of int
        Size of each dimension.
    inds : sequence of str
        Names of each dimension.
    tags : sequence of str
        Labels to tag this tensor with.
    dtype : {'float64', 'complex128', 'float32', 'complex64'}, optional
        The underlying data type.
    left_inds : sequence of str, optional
        Which, if any, indices to group as 'left' indices of an effective
        matrix. This can be useful, for example, when automatically applying
        unitary constraints to impose a certain flow on a tensor network but at
        the atomistic (Tensor) level.

    Returns
    -------
    Tensor
    """
    data = randn(shape, dtype=dtype)
    return Tensor(data=data, inds=inds, tags=tags, left_inds=left_inds)


@random_seed_fn
def rand_phased(shape, inds, tags=None, dtype=complex):
    """Generate a random tensor with specified shape and inds, and randomly
    'phased' (distributed on the unit circle) data, such that
    ``T.H @ T == T.norm()**2 == T.size``.

    Parameters
    ----------
    shape : sequence of int
        Size of each dimension.
    inds : sequence of str
        Names of each dimension.
    tags : sequence of str
        Labels to tag this tensor with.
    dtype : {'complex128', 'complex64'}, optional
        The underlying data type - can only be complex.

    Returns
    -------
    Tensor
    """
    data = rand_phase(shape, dtype=dtype)
    return Tensor(data=data, inds=inds, tags=tags)


# --------------------------------------------------------------------------- #
#                                    MPSs                                     #
# --------------------------------------------------------------------------- #

@random_seed_fn
def MPS_rand_state(n, bond_dim, phys_dim=2, normalize=True, cyclic=False,
                   dtype=float, trans_invar=False, **mps_opts):
    """Generate a random matrix product state.

    Parameters
    ----------
    n : int
        The number of sites.
    bond_dim : int
        The bond dimension.
    phys_dim : int, optional
        The physical (site) dimensions, defaults to 2.
    normalize : bool, optional
        Whether to normalize the state.
    cyclic : bool, optional
        Generate a MPS with periodic boundary conditions or not, default is
        open boundary conditions.
    dtype : {float, complex} or numpy dtype, optional
        Data type of the tensor network.
    trans_invar : bool (optional)
        Whether to generate a translationally invariant state,
        requires cyclic=True.
    mps_opts
        Supplied to :class:`~quimb.tensor.tensor_1d.MatrixProductState`.
    """
    if trans_invar and not cyclic:
        raise ValueError("State cannot be translationally invariant with open "
                         "boundary conditions.")

    cyc_dim = (bond_dim,) if cyclic else ()

    def gen_shapes():
        yield (*cyc_dim, bond_dim, phys_dim)
        for _ in range(n - 2):
            yield (bond_dim, bond_dim, phys_dim)
        yield (bond_dim, *cyc_dim, phys_dim)

    def gen_data(shape):
        return randn(shape, dtype=dtype)

    if trans_invar:
        array = sensibly_scale(gen_data(next(gen_shapes())))
        arrays = (array for _ in range(n))
    else:
        arrays = map(sensibly_scale, map(gen_data, gen_shapes()))

    rmps = MatrixProductState(arrays, **mps_opts)

    if normalize == 'left':
        rmps.left_canonize(normalize=True)
    elif normalize == 'right':
        rmps.left_canonize(normalize=True)
    elif normalize:
        rmps /= (rmps.H @ rmps)**0.5

    return rmps


def MPS_product_state(arrays, cyclic=False, **mps_opts):
    """Generate a product state in MatrixProductState form, i,e,
    with bond dimension 1, from single site vectors described by ``arrays``.
    """
    cyc_dim = (1,) if cyclic else ()

    def gen_array_shapes():
        yield (*cyc_dim, 1, -1)
        for _ in range(len(arrays) - 2):
            yield (1, 1, -1)
        yield (*cyc_dim, 1, -1)

    mps_arrays = (asarray(array).reshape(*shape)
                  for array, shape in zip(arrays, gen_array_shapes()))

    return MatrixProductState(mps_arrays, shape='lrp', **mps_opts)


def MPS_computational_state(binary, dtype=float, cyclic=False, **mps_opts):
    """A computational basis state in Matrix Product State form.

    Parameters
    ----------
    binary : str or sequence of int
        String specifying the state, e.g. ``'00101010111'`` or ``[0, 0, 1]``.
    cyclic : bool, optional
        Generate a MPS with periodic boundary conditions or not, default open
        boundary conditions.
    mps_opts
        Supplied to MatrixProductState constructor.
    """

    array_map = {
        '0': np.array([1., 0.], dtype=dtype),
        '1': np.array([0., 1.], dtype=dtype),
    }

    def gen_arrays():
        for s in binary:
            yield array_map[str(s)]

    return MPS_product_state(tuple(gen_arrays()), cyclic=cyclic, **mps_opts)


def MPS_neel_state(n, down_first=False, dtype=float, **mps_opts):
    """Generate the neel state in Matrix Product State form.

    Parameters
    ----------
    n : int
        The number of spins.
    down_first : bool, optional
        Whether to start with '1' (down) or '0' (up) first.
    mps_opts
        Supplied to MatrixProductState constructor.
    """
    binary_str = "01" * (n // 2) + (n % 2 == 1) * "0"
    if down_first:
        binary_str = "1" + binary_str[:-1]
    return MPS_computational_state(binary_str, dtype=dtype, **mps_opts)


def MPS_ghz_state(n, dtype=float, **mps_opts):
    """Build the chi=2 OBC MPS representation of the GHZ state.

    Parameters
    ----------
    n : int
        Number of qubits.
    dtype : {'float64', 'complex128', 'float32', 'complex64'}, optional
        The underlying data type.
    mps_opts
        Supplied to :class:`~quimb.tensor.tensor_1d.MatrixProductState`.
    """

    def gen_arrays():
        yield 2**-0.5 * np.array([[1., 0.],
                                  [0., 1.]]).astype(dtype)

        for i in range(1, n - 1):
            yield np.array([[[1., 0.],
                             [0., 0.]],
                            [[0., 0.],
                             [0., 1.]]]).astype(dtype)

        yield np.array([[1., 0.],
                        [0., 1.]]).astype(dtype)

    return MatrixProductState(gen_arrays(), **mps_opts)


def MPS_w_state(n, dtype=float, **mps_opts):
    """Build the chi=2 OBC MPS representation of the W state.

    Parameters
    ----------
    n : int
        Number of qubits.
    dtype : {'float64', 'complex128', 'float32', 'complex64'}, optional
        The underlying data type.
    mps_opts
        Supplied to :class:`~quimb.tensor.tensor_1d.MatrixProductState`.
    """

    def gen_arrays():
        yield (np.array([[1., 0.],
                         [0., 1.]]) / n ** 0.5).astype(dtype)

        for i in range(1, n - 1):
            yield np.array([[[1., 0.],
                             [0., 1.]],
                            [[0., 0.],
                             [1., 0.]]]).astype(dtype)

        yield np.array([[0., 1.],
                        [1., 0.]]).astype(dtype)

    return MatrixProductState(gen_arrays(), **mps_opts)


@random_seed_fn
def MPS_rand_computational_state(n, dtype=float, **mps_opts):
    """Generate a random computation basis state, like '01101001010'.

    Parameters
    ----------
    n : int
        The number of qubits.
    seed : int, optional
        The seed to use.
    dtype : {float, complex} or numpy dtype, optional
        Data type of the tensor network.
    mps_opts
        Supplied to :class:`~quimb.tensor.tensor_1d.MatrixProductState`.
    """
    cstr = (choice(('0', '1')) for _ in range(n))
    return MPS_computational_state(cstr, dtype=dtype, **mps_opts)


def MPS_zero_state(n, bond_dim=1, phys_dim=2, cyclic=False,
                   dtype=float, **mps_opts):
    """The all-zeros MPS state, of given bond-dimension.

    Parameters
    ----------
    n : int
        The number of sites.
    bond_dim : int, optional
        The bond dimension, defaults to 1.
    phys_dim : int, optional
        The physical (site) dimensions, defaults to 2.
    cyclic : bool, optional
        Generate a MPS with periodic boundary conditions or not, default is
        open boundary conditions.
    dtype : {float, complex} or numpy dtype, optional
        Data type of the tensor network.
    mps_opts
        Supplied to :class:`~quimb.tensor.tensor_1d.MatrixProductState`.
    """
    cyc_dim = (bond_dim,) if cyclic else ()

    def gen_arrays():
        yield np.zeros((*cyc_dim, bond_dim, phys_dim), dtype=dtype)
        for _ in range(n - 2):
            yield np.zeros((bond_dim, bond_dim, phys_dim), dtype=dtype)
        yield np.zeros((bond_dim, *cyc_dim, phys_dim), dtype=dtype)

    return MatrixProductState(gen_arrays(), **mps_opts)


def MPS_sampler(n, dtype=complex, squeeze=True, **mps_opts):
    """A product state for sampling tensor network traces. Seen as a vector it
    has the required property that ``psi.H @ psi == d`` always for hilbert
    space size ``d``.
    """
    arrays = [rand_phase(2, dtype=dtype) for _ in range(n)]
    psi = MPS_product_state(arrays, **mps_opts)
    if squeeze:
        psi.squeeze_()
    return psi


# --------------------------------------------------------------------------- #
#                                    MPOs                                     #
# --------------------------------------------------------------------------- #

def MPO_identity(n, phys_dim=2, dtype=float, cyclic=False, **mpo_opts):
    """Generate an identity MPO of size ``n``.

    Parameters
    ----------
    n : int
        The number of sites.
    phys_dim : int, optional
        The physical (site) dimensions, defaults to 2.
    dtype : {float, complex} or numpy dtype, optional
        Data type of the tensor network.
    cyclic : bool, optional
        Generate a MPO with periodic boundary conditions or not, default is
        open boundary conditions.
    mpo_opts
        Supplied to :class:`~quimb.tensor.tensor_1d.MatrixProductOperator`.
    """
    II = np.identity(phys_dim, dtype=dtype)
    cyc_dim = (1,) if cyclic else ()

    def gen_arrays():
        yield II.reshape(*cyc_dim, 1, phys_dim, phys_dim)
        for _ in range(n - 2):
            yield II.reshape(1, 1, phys_dim, phys_dim)
        yield II.reshape(1, *cyc_dim, phys_dim, phys_dim)

    return MatrixProductOperator(gen_arrays(), **mpo_opts)


def MPO_identity_like(mpo, **mpo_opts):
    """Return an identity matrix operator with the same physical index and
    inds/tags as ``mpo``.
    """
    return MPO_identity(n=mpo.nsites, phys_dim=mpo.phys_dim(), dtype=mpo.dtype,
                        site_tag_id=mpo.site_tag_id, cyclic=mpo.cyclic,
                        upper_ind_id=mpo.upper_ind_id,
                        lower_ind_id=mpo.lower_ind_id, **mpo_opts)


def MPO_zeros(n, phys_dim=2, dtype=float, cyclic=False, **mpo_opts):
    """Generate a zeros MPO of size ``n``.

    Parameters
    ----------
    n : int
        The number of sites.
    phys_dim : int, optional
        The physical (site) dimensions, defaults to 2.
    dtype : {float, complex} or numpy dtype, optional
        Data type of the tensor network.
    cyclic : bool, optional
        Generate a MPO with periodic boundary conditions or not, default is
        open boundary conditions.
    mpo_opts
        Supplied to :class:`~quimb.tensor.tensor_1d.MatrixProductOperator`.
    """
    cyc_dim = (1,) if cyclic else ()

    def gen_arrays():
        yield np.zeros((*cyc_dim, 1, phys_dim, phys_dim), dtype=dtype)
        for _ in range(n - 2):
            yield np.zeros((1, 1, phys_dim, phys_dim), dtype=dtype)
        yield np.zeros((1, *cyc_dim, phys_dim, phys_dim), dtype=dtype)

    return MatrixProductOperator(gen_arrays(), **mpo_opts)


def MPO_zeros_like(mpo, **mpo_opts):
    """Return a zeros matrix operator with the same physical index and
    inds/tags as ``mpo``.
    """
    return MPO_zeros(n=mpo.nsites, phys_dim=mpo.phys_dim(),
                     dtype=mpo.dtype, site_tag_id=mpo.site_tag_id,
                     upper_ind_id=mpo.upper_ind_id, cyclic=mpo.cyclic,
                     lower_ind_id=mpo.lower_ind_id, **mpo_opts)


@random_seed_fn
def MPO_rand(n, bond_dim, phys_dim=2, normalize=True, cyclic=False,
             herm=False, dtype=float, **mpo_opts):
    """Generate a random matrix product state.

    Parameters
    ----------
    n : int
        The number of sites.
    bond_dim : int
        The bond dimension.
    phys_dim : int, optional
        The physical (site) dimensions, defaults to 2.
    normalize : bool, optional
        Whether to normalize the operator such that ``trace(A.H @ A) == 1``.
    cyclic : bool, optional
        Generate a MPO with periodic boundary conditions or not, default is
        open boundary conditions.
    dtype : {float, complex} or numpy dtype, optional
        Data type of the tensor network.
    herm : bool, optional
        Whether to make the matrix hermitian (or symmetric if real) or not.
    mpo_opts
        Supplied to :class:`~quimb.tensor.tensor_1d.MatrixProductOperator`.
    """
    cyc_shp = (bond_dim,) if cyclic else ()

    shapes = [(*cyc_shp, bond_dim, phys_dim, phys_dim),
              *((bond_dim, bond_dim, phys_dim, phys_dim),) * (n - 2),
              (bond_dim, *cyc_shp, phys_dim, phys_dim)]

    def gen_data(shape):
        data = randn(shape, dtype=dtype)
        if not herm:
            return data

        trans = (0, 2, 1) if len(shape) == 3 else (0, 1, 3, 2)
        return data + data.transpose(*trans).conj()

    arrays = map(sensibly_scale, map(gen_data, shapes))

    rmpo = MatrixProductOperator(arrays, **mpo_opts)

    if normalize:
        rmpo /= (rmpo.H @ rmpo)**0.5

    return rmpo


@random_seed_fn
def MPO_rand_herm(n, bond_dim, phys_dim=2, normalize=True,
                  dtype=float, **mpo_opts):
    """Generate a random hermitian matrix product operator.
    See :class:`~quimb.tensor.tensor_gen.MPO_rand`.
    """
    return MPO_rand(n, bond_dim, phys_dim=phys_dim, normalize=normalize,
                    dtype=dtype, herm=True, **mpo_opts)


# ---------------------------- MPO hamiltonians ----------------------------- #

def maybe_make_real(X):
    """Check if ``X`` is real, if so, convert to contiguous array.
    """
    if np.allclose(X.imag, np.zeros_like(X)):
        return np.ascontiguousarray(X.real)
    return X


def spin_ham_mpo_tensor(one_site_terms, two_site_terms, S=1 / 2,
                        left_two_site_terms=None, which=None, cyclic=False):
    """Generate tensor(s) for a spin hamiltonian MPO.

    Parameters
    ----------
    one_site_terms : sequence of (scalar, operator)
        The terms that act on a single site, each ``operator`` can be a string
        suitable to be sent to :func:`spin_operator` or an actual 2d-array.
    two_site_terms : sequence of (scalar, operator operator)
        The terms that act on two neighbouring sites, each ``operator`` can be
        a string suitable to be sent to :func:`spin_operator` or an actual
        2d-array.
    S : fraction, optional
        What size spin to use, defaults to spin-1/2.
    left_two_site_terms : sequence of (scalar, operator operator), optional
        If the interaction to the left of this site has different spin terms
        then the equivalent list of terms for that site.
    which : {None, 'L', 'R', 'A'}, optional
        If ``None``, generate the middle tensor, if 'L' a left-end tensor, if
        'R' a right-end tensor and if 'A' all three.
    cyclic : bool, optional
        Whether to use periodic boundary conditions - default False.

    Returns
    -------
    numpy.ndarray[, numpy.ndarray, numpy.ndarray]
        The middle, left, right or all three MPO tensors.
    """
    # assume same interaction type everywhere
    if left_two_site_terms is None:
        left_two_site_terms = two_site_terms

    # local dimension
    D = int(2 * S + 1)
    # bond dimension to right
    B = len(two_site_terms) + 2
    # bond dimension to left
    BL = len(left_two_site_terms) + 2

    H = np.zeros((BL, B, D, D), dtype=complex)

    # add one-body terms
    for factor, s in one_site_terms:
        if isinstance(s, str):
            s = spin_operator(s, S=S)
        H[-1, 0, :, :] += factor * s

    # add two-body terms
    for i, (factor, s1, _) in enumerate(two_site_terms):
        if isinstance(s1, str):
            s1 = spin_operator(s1, S=S)
        H[-1, 1 + i, :, :] = factor * s1

    for i, (_, _, s2) in enumerate(left_two_site_terms):
        if isinstance(s2, str):
            s2 = spin_operator(s2, S=S)
        H[i + 1, 0, :, :] = s2

    H[0, 0, :, :] = eye(D)
    H[-1, -1, :, :] = eye(D)

    H = maybe_make_real(H)
    make_immutable(H)

    if which in {None, 'M'}:
        return H

    if cyclic:
        # need special conditions for first MPO matrix
        HL = np.zeros_like(H)
        HL[0, :, :, :] = H[-1, :, :, :]
        HL[1:-1, -1, :, :] = H[1:-1, 0, :, :]
        HR = H
    else:
        HL = H[-1, :, :, :]
        HR = H[:, 0, :, :]

    if which == 'L':
        return HL
    elif which == 'R':
        return HR
    elif which == 'A':
        return HL, H, HR


class _TermAdder:
    """Simple class to allow ``SpinHam`` syntax like
    ``builder[i, j] += (1/2, 'Z', 'X')``. This object is temporarily created
    by the getitem call, accumulates the new term, then has its the new
    combined list of terms extracted in the setitem call.
    """

    def __init__(self, terms, nsite):
        self.terms = terms
        self.nsite = nsite

    def __iadd__(self, new):
        if len(new) != self.nsite + 1:
            raise ValueError(
                "New terms should be of the form")

        if self.terms is None:
            self.terms = [new]
        else:
            self.terms += [new]
        return self


class SpinHam:
    """Class for easily building custom spin hamiltonians in MPO or NNI form.
    Currently limited to nearest neighbour interactions (and single site
    terms). It is possible to set 'default' translationally invariant terms,
    but also terms acting on specific sites only (which take precedence).
    It is also possible to build a sparse matrix version of the hamiltonian
    (obviously for small sizes only).

    Parameters
    ----------
    S : float, optional
        The type of spin, defaults to 1/2.
    cyclic : bool, optional
        Whether to use periodic boundary conditions - default is False.

    Examples
    --------
    Initialize the spin hamiltonian builder:

        >>> builder = SpinHam(S=3 / 2)

    Add some two-site terms:

        >>> builder += 0.5, '+', '-'
        >>> builder += 0.5, '-', '+'
        >>> builder += 1.0, 'Z', 'Z'

    Add a single site term:

        >>> builder -= 0.3, 'Z'

    Build a MPO version of the hamiltonian for use with DMRG:

        >>> mpo_ham = builder.build_mpo(100)
        >>> mpo_ham
        <MatrixProductOperator(tensors=100, structure='I{}', nsites=100)>

    Build a NNI version of the hamiltonian for use with TEBD:

        >>> builder.build_nni(100)
        <NNI(n=100, cyclic=False)>

    You can also set terms for specific sites (this overides any of the
    'default', translationally invariant terms set as above):

        >>> builder[10, 11] += 0.75, '+', '-'
        >>> builder[10, 11] += 0.75, '-', '+'
        >>> builder[10, 11] += 1.5, 'Z', 'Z'

    Or specific one-site terms (which again overides any default
    single site terms set above):

        >>> builder[10] += 3.7, 'Z'
        >>> builder[11] += 0.0, 'I' # '0' term turns off field
    """

    def __init__(self, S=1 / 2, cyclic=False):
        self.S = S
        self.one_site_terms = []
        self.two_site_terms = []
        self.cyclic = cyclic

        # Holders for any non-translationally invariant terms
        self.var_one_site_terms = {}
        self.var_two_site_terms = {}

    def add_term(self, factor, *operators):
        """Add another term to the expression to be built.

        Parameters
        ----------
        factor : scalar
            Scalar factor to multiply this term by.
        *operators : str or array
            The operators to use. Can specify one or two for single or two site
            terms respectively. Can use strings, which are supplied to
            :func:`~quimb.spin_operator`, or actual arrays as long as they have
            the correct dimension.
        """
        if factor == 0.0:
            # no need to add zero terms
            return

        if len(operators) == 1:
            self.one_site_terms.append((factor, *operators))
        elif len(operators) == 2:
            self.two_site_terms.append((factor, *operators))
        else:
            raise NotImplementedError("3-body+ terms are not supported yet.")

    def sub_term(self, factor, *operators):
        """Subtract a term - simple alias that flips sign of ``factor``.
        """
        self.add_term(-factor, *operators)

    def __iadd__(self, term):
        self.add_term(*term)
        return self

    def __isub__(self, term):
        self.sub_term(*term)
        return self

    def __getitem__(self, sites):
        """Part of the machinery that allows terms to be added to specific
        sites like::

            >>> builder[i] += 1/2, 'X'
            >>> builder[45, 46] += 1/2, 'Z', 'Z'

        """
        if isinstance(sites, Integral):
            return _TermAdder(self.var_one_site_terms.get(sites, None), 1)

        i, j = sorted(sites)
        if j - i != 1:
            raise NotImplementedError("Can only add nearest neighbour terms.")

        return _TermAdder(self.var_two_site_terms.get(sites, None), 2)

    def __setitem__(self, sites, value):
        """Part of the machinery that allows terms to be added to specific
        sites like::

            >>> builder[i] += 1/2, 'X'
            >>> builder[45, 46] += 1/2, 'Z', 'Z'

        Could also be called directly with a list of terms like::

            >>> builder[13, 14] = [(1, 'Z', 'Z'), (0.5, 'X', 'Y')]

        Which would overide any terms set so far.
        """
        if isinstance(value, _TermAdder):
            terms = value.terms
        else:
            terms = value

        if isinstance(sites, Integral):
            self.var_one_site_terms[sites] = terms
        else:
            i, j = sorted(sites)
            if j - i != 1:
                raise ValueError("Can only add nearest neighbour terms.")
            self.var_two_site_terms[sites] = terms

    def build_mpo(self, n, upper_ind_id='k{}', lower_ind_id='b{}',
                  site_tag_id='I{}', tags=None, bond_name=""):
        """Build an MPO instance of this spin hamiltonian of size ``n``. See
        also ``MatrixProductOperator``.
        """
        # cache the default term
        t_defs = {}

        def get_default_term(which):
            try:
                return t_defs[which]
            except KeyError:
                t_defs['L'], t_defs[None], t_defs['R'] = spin_ham_mpo_tensor(
                    self.one_site_terms, self.two_site_terms,
                    S=self.S, which='A', cyclic=self.cyclic)
                return t_defs[which]

        def gen_tensors():
            for i in range(n):
                which = {0: 'L', n - 1: 'R'}.get(i, None)

                ij_L = (i - 1, i)
                ij_R = (i, i + 1)

                # check for site/bond specific terms
                var_one = i in self.var_one_site_terms
                var_two = (
                    (ij_L in self.var_two_site_terms) or
                    (ij_R in self.var_two_site_terms)
                )

                if not (var_one or var_two):
                    yield get_default_term(which)
                else:
                    t1s = self.var_one_site_terms.get(i, self.one_site_terms)
                    t2s = self.var_two_site_terms.get(ij_R,
                                                      self.two_site_terms)
                    t2s_L = self.var_two_site_terms.get(ij_L,
                                                        self.two_site_terms)

                    yield spin_ham_mpo_tensor(t1s, t2s, S=self.S,
                                              left_two_site_terms=t2s_L,
                                              which=which, cyclic=self.cyclic)

        return MatrixProductOperator(arrays=gen_tensors(), bond_name=bond_name,
                                     upper_ind_id=upper_ind_id,
                                     lower_ind_id=lower_ind_id,
                                     site_tag_id=site_tag_id, tags=tags)

    def build_sparse(self, n, **ikron_opts):
        """Build a sparse matrix representation of this Hamiltonian.

        Parameters
        ----------
        n : int, optional
            The number of spins to build the matrix for.
        ikron_opts
            Supplied to :func:`~quimb.core.ikron`.

        Returns
        -------
        H : matrix
        """
        ikron_opts.setdefault('sparse', True)

        D = int(2 * self.S + 1)
        dims = [D] * n

        terms = []
        for i in range(n):

            t1s = self.var_one_site_terms.get(i, self.one_site_terms)
            for factor, s in t1s:
                if isinstance(s, str):
                    s = spin_operator(s, S=self.S, sparse=True)
                terms.append(
                    ikron(factor * s, dims, i, **ikron_opts)
                )

            if (i + 1 == n) and (not self.cyclic):
                break

            t2s = self.var_two_site_terms.get((i, i + 1), self.two_site_terms)
            for factor, s1, s2 in t2s:
                if isinstance(s1, str):
                    s1 = spin_operator(s1, S=self.S, sparse=True)
                if isinstance(s2, str):
                    s2 = spin_operator(s2, S=self.S, sparse=True)
                terms.append(
                    ikron([factor * s1, s2], dims, [i, i + 1], **ikron_opts)
                )

        return sum(terms)

    def _get_spin_op(self, factor, *ss):
        if len(ss) == 1:
            s, = ss
            if isinstance(s, str):
                s = spin_operator(s, S=self.S)
            return factor * s

        if len(ss) == 2:
            s1, s2 = ss
            if isinstance(s1, str):
                s1 = spin_operator(s1, S=self.S)
            if isinstance(s2, str):
                s2 = spin_operator(s2, S=self.S)
            return factor * (s1 & s2)

    def _sum_spin_ops(self, terms):
        H = sum(self._get_spin_op(*term) for term in terms)
        H = maybe_make_real(H)
        make_immutable(H)
        return H

    def build_nni(self, n=None, **nni_opts):
        """Build a nearest neighbour interactor instance of this spin
        hamiltonian of size ``n``. See also
        :class:`~quimb.tensor.tensor_tebd.NNI`.

        Parameters
        ----------
        n : int, optional
            The number of spins, if the hamiltonian only has two-site terms
            this is optional.

        Returns
        -------
        NNI
        """
        H1s, H2s = {}, {}

        # add default two site term
        if self.two_site_terms:
            H2s[None] = self._sum_spin_ops(self.two_site_terms)

        # add specific two site terms
        if self.var_two_site_terms:
            for sites, terms in self.var_two_site_terms.items():
                H2s[sites] = self._sum_spin_ops(terms)

        # add default one site term
        if self.one_site_terms:
            H1s[None] = self._sum_spin_ops(self.one_site_terms)

        # add specific one site terms
        if self.var_one_site_terms:
            for site, terms in self.var_one_site_terms.items():
                H1s[site] = self._sum_spin_ops(terms)

        return NNI(H2=H2s, H1=H1s, n=n, cyclic=self.cyclic, **nni_opts)


def _ham_ising(j=1.0, bx=0.0, *, S=1 / 2, cyclic=False):
    H = SpinHam(S=1 / 2, cyclic=cyclic)
    H += j, 'Z', 'Z'
    H -= bx, 'X'
    return H


def MPO_ham_ising(n, j=1.0, bx=0.0, *, S=1 / 2, cyclic=False, **mpo_opts):
    """Ising Hamiltonian in MPO form.

    Parameters
    ----------
    n : int
        The number of sites.
    j : float, optional
        The ZZ interaction strength.
    bx : float, optional
        The X-magnetic field strength.
    S : {1/2, 1, 3/2, ...}, optional
        The underlying spin of the system, defaults to 1/2.
    cyclic : bool, optional
        Generate a MPO with periodic boundary conditions or not, default is
        open boundary conditions.
    mpo_opts or nni_opts
        Supplied to :class:`~quimb.tensor.tensor_1d.MatrixProductOperator`.

    Returns
    -------
    MatrixProductOperator
    """
    H = _ham_ising(j=j, bx=bx, S=S, cyclic=cyclic)
    return H.build_mpo(n, **mpo_opts)


def NNI_ham_ising(n=None, j=1.0, bx=0.0, *, S=1 / 2, cyclic=False, **nni_opts):
    """Ising Hamiltonian in NNI form.

    Parameters
    ----------
    n : int
        The number of sites.
    j : float, optional
        The ZZ interaction strength.
    bx : float, optional
        The X-magnetic field strength.
    S : {1/2, 1, 3/2, ...}, optional
        The underlying spin of the system, defaults to 1/2.
    cyclic : bool, optional
        Generate a NNI with periodic boundary conditions or not, default is
        open boundary conditions.
    mpo_opts or nni_opts
        Supplied to :class:`~quimb.tensor.tensor_1d.NNI`.

    Returns
    -------
    NNI
    """
    H = _ham_ising(j=j, bx=bx, S=S, cyclic=cyclic)
    return H.build_nni(n=n, **nni_opts)


def _ham_XY(j=1.0, bz=0.0, *, S=1 / 2, cyclic=False):
    H = SpinHam(S=S, cyclic=cyclic)

    try:
        jx, jy = j
    except (TypeError, ValueError):
        jx = jy = j

    if jx == jy:
        # easy way to enforce realness
        H += jx / 2, '+', '-'
        H += jx / 2, '-', '+'
    else:
        H += jx, 'X', 'X'
        H += jy, 'Y', 'Y'

    H -= bz, 'Z'

    return H


def MPO_ham_XY(n, j=1.0, bz=0.0, *, S=1 / 2, cyclic=False, **mpo_opts):
    """XY-Hamiltonian in MPO form.

    Parameters
    ----------
    n : int
        The number of sites.
    j : float or (float, float), optional
        The XX and YY interaction strength.
    bz : float, optional
        The Z-magnetic field strength.
    S : {1/2, 1, 3/2, ...}, optional
        The underlying spin of the system, defaults to 1/2.
    cyclic : bool, optional
        Generate a MPO with periodic boundary conditions or not, default is
        open boundary conditions.
    mpo_opts or nni_opts
        Supplied to :class:`~quimb.tensor.tensor_1d.MatrixProductOperator`.

    Returns
    -------
    MatrixProductOperator
    """
    H = _ham_XY(j=j, bz=bz, S=S, cyclic=cyclic)
    return H.build_mpo(n, **mpo_opts)


def NNI_ham_XY(n=None, j=1.0, bz=0.0, *, S=1 / 2, cyclic=False, **nni_opts):
    """XY-Hamiltonian in NNI form.

    Parameters
    ----------
    n : int
        The number of sites.
    j : float or (float, float), optional
        The XX and YY interaction strength.
    bz : float, optional
        The Z-magnetic field strength.
    S : {1/2, 1, 3/2, ...}, optional
        The underlying spin of the system, defaults to 1/2.
    cyclic : bool, optional
        Generate a NNI with periodic boundary conditions or not, default is
        open boundary conditions.
    nni_opts
        Supplied to :class:`~quimb.tensor.tensor_1d.NNI`.

    Returns
    -------
    NNI
    """
    H = _ham_XY(j=j, bz=bz, S=S, cyclic=cyclic)
    return H.build_nni(n=n, **nni_opts)


def _ham_heis(j=1.0, bz=0.0, *, S=1 / 2, cyclic=False):
    H = SpinHam(S=S, cyclic=cyclic)

    try:
        jx, jy, jz = j
    except (TypeError, ValueError):
        jx = jy = jz = j

    if jx == jy:
        # easy way to enforce realness
        H += jx / 2, '+', '-'
        H += jx / 2, '-', '+'
    else:
        H += jx, 'X', 'X'
        H += jy, 'Y', 'Y'
    H += jz, 'Z', 'Z'

    H -= bz, 'Z'

    return H


def MPO_ham_heis(n, j=1.0, bz=0.0, *, S=1 / 2, cyclic=False, **mpo_opts):
    """Heisenberg Hamiltonian in MPO form.

    Parameters
    ----------
    n : int
        The number of sites.
    j : float or (float, float, float), optional
        The XX, YY and ZZ interaction strength.
    bz : float, optional
        The Z-magnetic field strength.
    S : {1/2, 1, 3/2, ...}, optional
        The underlying spin of the system, defaults to 1/2.
    cyclic : bool, optional
        Generate a MPO with periodic boundary conditions or not, default is
        open boundary conditions.
    mpo_opts
        Supplied to :class:`~quimb.tensor.tensor_1d.MatrixProductOperator`.

    Returns
    -------
    MatrixProductOperator
    """
    H = _ham_heis(j=j, bz=bz, S=S, cyclic=cyclic)
    return H.build_mpo(n, **mpo_opts)


def NNI_ham_heis(n=None, j=1.0, bz=0.0, *, S=1 / 2, cyclic=False, **nni_opts):
    """Heisenberg Hamiltonian in NNI form.

    Parameters
    ----------
    n : int
        The number of sites.
    j : float or (float, float, float), optional
        The XX, YY and ZZ interaction strength.
    bz : float, optional
        The Z-magnetic field strength.
    S : {1/2, 1, 3/2, ...}, optional
        The underlying spin of the system, defaults to 1/2.
    cyclic : bool, optional
        Generate a NNI with periodic boundary conditions or not, default is
        open boundary conditions.
    nni_opts
        Supplied to :class:`~quimb.tensor.tensor_gen.NNI`.

    Returns
    -------
    NNI
    """
    H = _ham_heis(j=j, bz=bz, S=S, cyclic=cyclic)
    return H.build_nni(n=n, **nni_opts)


def MPO_ham_XXZ(n, delta, jxy=1.0, *, S=1 / 2, cyclic=False, **mpo_opts):
    """XXZ-Hamiltonian in MPO form.

    Parameters
    ----------
    n : int
        The number of sites.
    delta : float
        The Z-interaction strength.
    jxy : float, optional
        The X- and Y- interaction strength, defaults to 1.
    S : {1/2, 1, 3/2, ...}, optional
        The underlying spin of the system, defaults to 1/2.
    cyclic : bool, optional
        Generate a MPO with periodic boundary conditions or not, default is
        open boundary conditions.
    mpo_opts
        Supplied to :class:`~quimb.tensor.tensor_1d.MatrixProductOperator`.

    Returns
    -------
    MatrixProductOperator
    """
    return MPO_ham_heis(n, j=(jxy, jxy, delta), S=S, cyclic=cyclic, **mpo_opts)


def NNI_ham_XXZ(n=None, delta=None, jxy=1.0, *,
                S=1 / 2, cyclic=False, **nni_opts):
    """XXZ-Hamiltonian in NNI form.

    Parameters
    ----------
    n : int
        The number of sites.
    delta : float
        The Z-interaction strength.
    jxy : float, optional
        The X- and Y- interaction strength, defaults to 1.
    S : {1/2, 1, 3/2, ...}, optional
        The underlying spin of the system, defaults to 1/2.
    cyclic : bool, optional
        Generate a NNI with periodic boundary conditions or not, default is
        open boundary conditions.
    nni_opts
        Supplied to :class:`~quimb.tensor.tensor_gen.NNI`.

    Returns
    -------
    NNI
    """
    if delta is None:
        raise ValueError("You need to specify ``delta``.")
    return NNI_ham_heis(n, j=(jxy, jxy, delta), S=S, cyclic=cyclic, **nni_opts)


def _ham_bilinear_biquadratic(theta, *, S=1 / 2, cyclic=False):
    H = SpinHam(S=S, cyclic=cyclic)

    H += np.cos(theta), 'X', 'X'
    H += np.cos(theta), 'Y', 'Y'
    H += np.cos(theta), 'Z', 'Z'

    XX = spin_operator('X', S=S) @ spin_operator('X', S=S)
    YY = spin_operator('Y', S=S) @ spin_operator('Y', S=S)
    ZZ = spin_operator('Z', S=S) @ spin_operator('Z', S=S)

    H += np.sin(theta), XX, XX
    H += np.sin(theta), XX, YY
    H += np.sin(theta), XX, ZZ
    H += np.sin(theta), YY, XX
    H += np.sin(theta), YY, YY
    H += np.sin(theta), YY, ZZ
    H += np.sin(theta), ZZ, XX
    H += np.sin(theta), ZZ, YY
    H += np.sin(theta), ZZ, ZZ

    return H


def MPO_ham_bilinear_biquadratic(n=None, theta=0, *, S=1 / 2, cyclic=False,
                                 compress=True, **mpo_opts):
    """ Hamiltonian of one-dimensional bilinear biquadratic chain in MPO form,
    see PhysRevB.93.184428.

    Parameters
    ----------
    n : int
        The number of sites.
    theta : float or (float, float), optional
        The parameter for linear and non-linear term of interaction strength,
        defaults to 0.
    S : {1/2, 1, 3/2, ...}, optional
        The underlying spin of the system, defaults to 1/2.
    cyclic : bool, optional
        Generate a NNI with periodic boundary conditions or not, default is
        open boundary conditions.
    mpo_opts
        Supplied to :class:`~quimb.tensor.tensor_1d.MatrixProductOperator`.

    Returns
    -------
    MatrixProductOperator
    """
    H = _ham_bilinear_biquadratic(theta, S=S, cyclic=cyclic)
    H_mpo = H.build_mpo(n, **mpo_opts)
    if compress is True:
        H_mpo.compress(cutoff=1e-12, cutoff_mode='rel' if cyclic else 'sum2')
    return H_mpo


def NNI_ham_bilinear_biquadratic(n=None, theta=0, *, S=1 / 2,
                                 cyclic=False, **nni_opts):
    """ Hamiltonian of one-dimensional bilinear biquadratic chain in NNI form,
    see PhysRevB.93.184428.

    Parameters
    ----------
    n : int
        The number of sites.
    theta : float or (float, float), optional
        The parameter for linear and non-linear term of interaction strength,
        defaults to 0.
    S : {1/2, 1, 3/2, ...}, optional
        The underlying spin of the system, defaults to 1/2.
    cyclic : bool, optional
        Generate a NNI with periodic boundary conditions or not, default is
        open boundary conditions.
    nni_opts
        Supplied to :class:`~quimb.tensor.tensor_gen.NNI`.

    Returns
    -------
    NNI
    """
    H = _ham_bilinear_biquadratic(theta, S=S, cyclic=cyclic)
    return H.build_nni(n=n, **nni_opts)


def _ham_mbl(n, dh, j=1.0, seed=None, S=1 / 2, *, cyclic=False,
             dh_dist='s', dh_dim=1, beta=None):
    # start with the heisenberg builder
    H = _ham_heis(j, S=S, cyclic=cyclic)

    dhds, rs = _gen_mbl_random_factors(n, dh, dh_dim, dh_dist, seed, beta)

    # generate noise, potentially in all directions, each with own strength
    for i in range(n):
        dh_r_xyzs = zip(dhds, rs[:, i], 'XYZ')
        for dh, r, xyz in dh_r_xyzs:
            H[i] += dh * r, xyz

    return H


def MPO_ham_mbl(n, dh, j=1.0, seed=None, S=1 / 2, *, cyclic=False,
                dh_dist='s', dh_dim=1, beta=None, **mpo_opts):
    """The many-body-localized spin hamiltonian in MPO form.

    Parameters
    ----------
    n : int
        Number of spins.
    dh : float
        Random noise strength.
    j : float, or (float, float, float), optional
        Interaction strength(s) e.g. 1 or (1., 1., 0.5).
    seed : int, optional
        Random number to seed the noise with.
    S : {1/2, 1, 3/2, ...}, optional
        The underlying spin of the system, defaults to 1/2.
    cyclic : bool, optional
        Whether to use periodic boundary conditions - default is False.
    dh_dist : {'s', 'g', 'qp'}, optional
        Whether to use sqaure, guassian or quasiperiodic noise.
    beta : float, optional
        Frequency of the quasirandom noise, only if ``dh_dist='qr'``.
    mpo_opts
        Supplied to :class:`MatrixProductOperator`.

    Returns
    -------
    MatrixProductOperator
    """
    H = _ham_mbl(n, dh=dh, j=j, seed=seed, S=S, cyclic=cyclic,
                 dh_dist=dh_dist, dh_dim=dh_dim, beta=beta)
    return H.build_mpo(n, **mpo_opts)


def NNI_ham_mbl(n, dh, j=1.0, seed=None, S=1 / 2, *, cyclic=False,
                dh_dist='s', dh_dim=1, beta=None, **nni_opts):
    """The many-body-localized spin hamiltonian in NNI form.

    Parameters
    ----------
    n : int
        Number of spins.
    dh : float
        Random noise strength.
    j : float, or (float, float, float), optional
        Interaction strength(s) e.g. 1 or (1., 1., 0.5).
    seed : int, optional
        Random number to seed the noise with.
    S : {1/2, 1, 3/2, ...}, optional
        The underlying spin of the system, defaults to 1/2.
    cyclic : bool, optional
        Whether to use periodic boundary conditions - default is False.
    dh_dist : {'s', 'g', 'qp'}, optional
        Whether to use sqaure, guassian or quasiperiodic noise.
    beta : float, optional
        Frequency of the quasirandom noise, only if ``dh_dist='qr'``.
    nni_opts
        Supplied to :class:`NNI`.

    Returns
    -------
    NNI
    """
    H = _ham_mbl(n, dh=dh, j=j, seed=seed, S=S, cyclic=cyclic,
                 dh_dist=dh_dist, dh_dim=dh_dim, beta=beta)
    return H.build_nni(n, **nni_opts)



import numpy as np

import quimb as qu


class NNI:
    """An simple interacting hamiltonian object used, for instance, in TEBD.
    Once instantiated, the ``NNI`` hamiltonian can be called like ``H_nni()``
    to get the default two-site term, or ``H_nni((i, j))`` to get the term
    specific to sites ``i`` and ``j``.

    If the terms supplied are anything but a single, two-site term, then the
    length of hamiltonian ``n`` must be specified too as the gates will no
    longer be completely translationally invariant.

    Parameters
    ----------
    H2 : array_like or dict[tuple[int], array_like]
        The sum of interaction terms. If a dict is given, the keys should be
        nearest neighbours like ``(10, 11)``, apart from any default term which
        should have the key ``None``, and the values should be the sum of
        interaction terms for that interaction.
    H1 : array_like or dict[int, array_like], optional
        The sum of single site terms. If a dict is given, the keys should be
        integer sites, apart from any default term which should have the key
        ``None``, and the values should be the sum of single site terms for
        that site.
    n : int, optional
        The size of the hamiltonian.
    cyclic : bool, optional
        Whether the hamiltonian has periodic boundary conditions or not.

    Attributes
    ----------
    special_sites : set[(int, int)]
        This keeps track of which pairs of sites don't just have the default
        term

    Examples
    --------
    A simple, translationally invariant, interaction-only ``NNI``::

        >>> XX = pauli('X') & pauli('X')
        >>> YY = pauli('Y') & pauli('Y')
        >>> H_nni = NNI(XX + YY)

    The same, but with a translationally invariant field as well (need to set
    ``n`` since the last gate will be different)::

        >>> Z = pauli('Z')
        >>> H_nni = NNI(H2=XX + YY, H1=Z, n=100)

    Specifying a default interaction and field, with custom values set for some
    sites::

        >>> H2 = {None: XX + YY, (49, 50): (XX + YY) / 2}
        >>> H1 = {None: Z, 49: 2 * Z, 50: 2 * Z}
        >>> H_nni = NNI(H2=H2, H1=H1, n=100)

    Specifying the hamiltonian entirely through site specific interactions and
    fields::

        >>> H2 = {(i, i + 1): XX + YY for i in range(99)}
        >>> H1 = {i: Z for i in range(100)}
        >>> H_nni = NNI(H2=H2, H1=H1, n=100)

    See Also
    --------
    SpinHam
    """

    def __init__(self, H2, H1=None, n=None, cyclic=False):
        self.n = n
        self.cyclic = cyclic

        if isinstance(H2, np.ndarray):
            H2 = {None: H2}
        if isinstance(H1, np.ndarray):
            H1 = {None: H1}

        self.H2s = dict(H2)
        self.H2s.setdefault(None, None)

        if H1 is not None:
            self.H1s = dict(H1)
        else:
            self.H1s = {}
        self.H1s.setdefault(None, None)

        # sites where the term might be different
        self.special_sites = {ij for ij in self.H2s if ij is not None}
        self.special_sites |= {(i, i + 1) for i in self.H1s if i is not None}
        obc_with_field = (not self.cyclic) and (self.H1s[None] is not None)

        # make sure n is supplied if it is needed
        if n is None:
            if (self.special_sites or obc_with_field):
                raise ValueError("Need to specify ``n`` if this ``NNI`` is "
                                 "anything but completely translationally "
                                 "invariant (including OBC w/ field).")

        # manually add the last interaction as a special site for OBC w/ field
        #     since the last gate has to apply single site field to both sites
        elif not self.cyclic:
            if obc_with_field or (self.n - 1 in self.H1s):
                self.special_sites.add((self.n - 2, self.n - 1))

        # this is the cache for holding generated two-body terms
        self._terms = {}

    def gen_term(self, sites=None):
        """Generate the interaction term acting on ``sites``.
        """
        # make sure have sites as (i, i + 1) if supplied
        if sites is not None:
            i, j = sites = tuple(sorted(sites))
            if j - i != 1:
                raise ValueError("Only nearest neighbour interactions are "
                                 "supported for an ``NNI``.")
        else:
            i = j = None

        term = self.H2s.get(sites, self.H2s[None])
        if term is None:
            raise ValueError(f"No term has been set for sites {sites}, either "
                             "specifically or via a default term.")

        # add single site term to left site if present
        H1 = self.H1s.get(i, self.H1s[None])

        # but only if this site has a term set
        if H1 is not None:
            I_2 = qu.eye(H1.shape[0], dtype=H1.dtype)
            term = term + qu.kron(H1, I_2)

        # if not PBC, for the last interaction, add to right site as well
        if sites and (j == self.n - 1) and (not self.cyclic):
            H1 = self.H1s.get(j, self.H1s[None])

            # but again, only if that site has a term set
            if H1 is not None:
                I_2 = qu.eye(H1.shape[0], dtype=H1.dtype)
                term = term + qu.kron(I_2, H1)

        return term

    def __call__(self, sites=None):
        """Get the cached term for sites ``sites``, generate if necessary.
        """
        try:
            return self._terms[sites]
        except KeyError:
            term = self.gen_term(sites)
            self._terms[sites] = term
            return term

    def mean_norm(self, ntype='fro'):
        """Computes the average frobenius norm of local terms. Also generates
        all terms if not already cached.
        """
        if self.n is None:
            return qu.norm(self(), ntype)

        nterms = self.n - int(not self.cyclic)
        return sum(
            qu.norm(self((i, i + 1)), ntype)
            for i in range(nterms)
        ) / nterms

    def __repr__(self):
        return f"<NNI(n={self.n}, cyclic={self.cyclic})>"


class TEBD:
    """Class implementing Time Evolving Block Decimation (TEBD) [1].

    [1] Guifré Vidal, Efficient Classical Simulation of Slightly Entangled
    Quantum Computations, PRL 91, 147902 (2003)

    Parameters
    ----------
    p0 : MatrixProductState
        Initial state.
    H : NNI or array_like
        Dense hamiltonian representing the two body interaction. Should have
        shape ``(d * d, d * d)``, where ``d`` is the physical dimension of
        ``p0``.
    dt : float, optional
        Default time step, cannot be set as well as ``tol``.
    tol : float, optional
        Default target error for each evolution, cannot be set as well as
        ``dt``, which will instead be calculated from the trotter orderm length
        of time, and hamiltonian norm.
    t0 : float, optional
        Initial time. Defaults to 0.0.
    split_opts : dict, optional
        Compression options applied for splitting after gate application, see
        :func:`~quimb.tensor.tensor_core.tensor_split`.

    See Also
    --------
    quimb.Evolution
    """

    def __init__(self, p0, H, dt=None, tol=None, t0=0.0,
                 split_opts=None, progbar=True):
        # prepare initial state
        self._pt = p0.copy()
        self._pt.canonize(0)
        self.N = self._pt.nsites

        # handle hamiltonian -> convert array to NNI
        if isinstance(H, np.ndarray):
            H = NNI(H, cyclic=p0.cyclic)
        if not isinstance(H, NNI):
            raise TypeError("``H`` should be a ``NNI`` or 2-site array, "
                            "not a TensorNetwork of any form.")

        if p0.cyclic != H.cyclic:
            raise ValueError("Both ``p0`` and ``H`` should have matching OBC "
                             "or PBC.")

        self.H = H
        self.cyclic = H.cyclic
        self._ham_norm = H.mean_norm()
        self._U_ints = {}
        self._err = 0.0

        # set time and tolerance defaults
        self.t0 = self.t = t0
        if dt and tol:
            raise ValueError("Can't set default for both ``dt`` and ``tol``.")
        self.dt = self._dt = dt
        self.tol = tol

        # misc other options
        self.progbar = progbar
        self.split_opts = {} if split_opts is None else dict(split_opts)

    @property
    def pt(self):
        """The MPS state of the system at the current time.
        """
        return self._pt.copy()

    @property
    def err(self):
        return self._err

    def choose_time_step(self, tol, T, order):
        """Trotter error is ``~ (T / dt) * dt^(order + 1)``. Invert to
        find desired time step, and scale by norm of interaction term.
        """
        return (tol / (T * self._ham_norm)) ** (1 / order)

    def get_gate(self, dt_frac, sites=None):
        """Get the unitary (exponentiated) gate for fraction of timestep
        ``dt_frac`` and sites ``sites``, cached.
        """
        if sites not in self.H.special_sites:
            sites = None

        try:
            return self._U_ints[dt_frac, sites]
        except KeyError:
            U = qu.expm(-1.0j * self._dt * dt_frac * self.H(sites))
            self._U_ints[dt_frac, sites] = U
            return U

    def sweep(self, direction, dt_frac, dt=None, queue=False):
        """Perform a single sweep of gates and compression. This shifts the
        orthonognality centre along with the gates as they are applied and
        split.

        Parameters
        ----------
        direction : {'right', 'left'}
            Which direction to sweep. Right is even bonds, left is odd.
        dt_frac : float
            What fraction of dt substep to take.
        dt : float, optional
            Overide the current ``dt`` with a custom value.
        """

        # if custom dt set, scale the dt fraction
        if dt is not None:
            dt_frac *= (dt / self._dt)

        # ------ automatically combine consecutive sweeps of same time ------ #

        if not hasattr(self, '_queued_sweep'):
            self._queued_sweep = None

        if queue:
            # check for queued sweep
            if self._queued_sweep:
                # if matches, combine and continue
                if direction == self._queued_sweep[0]:
                    self._queued_sweep[1] += dt_frac
                    return
                # else perform the old, queue the new
                else:
                    new_queued_sweep = [direction, dt_frac]
                    direction, dt_frac = self._queued_sweep
                    self._queued_sweep = new_queued_sweep

            # just queue the new sweep
            else:
                self._queued_sweep = [direction, dt_frac]
                return

        # check if need to drain the queue first
        elif self._queued_sweep:
            queued_direction, queued_dt_frac = self._queued_sweep
            self._queued_sweep = None
            self.sweep(queued_direction, queued_dt_frac, queue=False)

        # ------------------------------------------------------------------- #

        if direction == 'right':
            # Apply even gates:
            #
            #     o-<-<-<-<-<-<-<-<-<-   -<-<
            #     | | | | | | | | | |     | |       >~>~>~>~>~>~>~>~>~>~>~o
            #     UUU UUU UUU UUU UUU ... UUU  -->  | | | | | | | | | | | |
            #     | | | | | | | | | |     | |
            #      1   2   3   4   5  ==>
            #
            for i in range(0, self.N - 1, 2):
                sites = (i, i + 1)
                U = self.get_gate(dt_frac, sites)
                self._pt.left_canonize(start=max(0, i - 1), stop=i)
                self._pt.gate_split_(
                    U, where=sites, absorb='right', **self.split_opts)

        elif direction == 'left':
            # Apply odd gates:
            #
            #     >->->-   ->->->->->->->->-o
            #     | | |     | | | | | | | | |       o~<~<~<~<~<~<~<~<~<~<~<
            #     | UUU ... UUU UUU UUU UUU |  -->  | | | | | | | | | | | |
            #     | | |     | | | | | | | | |
            #           <==  4   3   2   1
            #
            for i in reversed(range(1, self.N - (0 if self.cyclic else 1), 2)):
                sites = (i, i + 1)
                U = self.get_gate(dt_frac, sites)
                self._pt.right_canonize(
                    start=min(self.N - 1, i + 2), stop=i + 1)
                self._pt.gate_split_(
                    U, where=sites, absorb='left', **self.split_opts)

            # one extra canonicalization not included in last split
            self._pt.right_canonize_site(1)

    def _step_order2(self, tau=1, **sweep_opts):
        """Perform a single, second order step.
        """
        self.sweep('right', tau / 2, **sweep_opts)
        self.sweep('left', tau, **sweep_opts)
        self.sweep('right', tau / 2, **sweep_opts)

    def _step_order4(self, **sweep_opts):
        """Perform a single, fourth order step.
        """
        tau1 = tau2 = 1 / (4 * 4**(1 / 3))
        tau3 = 1 - 2 * tau1 - 2 * tau2
        self._step_order2(tau1, **sweep_opts)
        self._step_order2(tau2, **sweep_opts)
        self._step_order2(tau3, **sweep_opts)
        self._step_order2(tau2, **sweep_opts)
        self._step_order2(tau1, **sweep_opts)

    def step(self, order=2, dt=None, progbar=None, **sweep_opts):
        """Perform a single step of time ``self.dt``.
        """
        {2: self._step_order2,
         4: self._step_order4}[order](dt=dt, **sweep_opts)

        dt = self._dt if dt is None else dt
        self.t += dt
        self._err += self._ham_norm * dt ** (order + 1)

        if progbar is not None:
            progbar.cupdate(self.t)
            self._set_progbar_desc(progbar)

    def _compute_sweep_dt_tol(self, T, dt, tol, order):
        # Work out timestep, possibly from target tol, and checking defaults
        dt = self.dt if (dt is None) else dt
        tol = self.tol if (tol is None) else tol

        if not (dt or tol):
            raise ValueError("Must set one of ``dt`` and ``tol``.")
        if (dt and tol):
            raise ValueError("Can't set both ``dt`` and ``tol``.")

        if dt is None:
            self._dt = self.choose_time_step(tol, T - self.t, order)
        else:
            self._dt = dt

        return self._dt

    TARGET_TOL = 1e-13  # tolerance to have 'reached' target time

    def update_to(self, T, dt=None, tol=None, order=4, progbar=None):
        """Update the state to time ``T``.

        Parameters
        ----------
        T : float
            The time to evolve to.
        dt : float, optional
            Time step to use. Can't be set as well as ``tol``.
        tol : float, optional
            Tolerance for whole evolution. Can't be set as well as ``dt``.
        order : int, optional
            Trotter order to use.
        progbar : bool, optional
            Manually turn the progress bar off.
        """
        if T < self.t - self.TARGET_TOL:
            raise NotImplementedError

        self._compute_sweep_dt_tol(T, dt, tol, order)

        # set up progress bar and start evolution
        progbar = self.progbar if (progbar is None) else progbar
        progbar = qu.utils.continuous_progbar(self.t, T) if progbar else None

        while self.t < T - self.TARGET_TOL:
            if (T - self.t < self._dt):
                # set custom dt if within one step of final time
                dt = T - self.t
                # also make sure queued sweeps are drained
                queue = False
            else:
                dt = None
                queue = True

            # perform a step!
            self.step(order=order, progbar=progbar, dt=dt, queue=queue)

        if progbar:
            progbar.close()

    def _set_progbar_desc(self, progbar):
        msg = f"t={self.t:.4g}, max-bond={self._pt.max_bond()}"
        progbar.set_description(msg)

    def at_times(self, ts, dt=None, tol=None, order=4, progbar=None):
        """Generate the time evolved state at each time in ``ts``.

        Parameters
        ----------
        ts : sequence of float
            The times to evolve to and yield the state at.
        dt : float, optional
            Time step to use. Can't be set as well as ``tol``.
        tol : float, optional
            Tolerance for whole evolution. Can't be set as well as ``dt``.
        order : int, optional
            Trotter order to use.
        progbar : bool, optional
            Manually turn the progress bar off.

        Yields
        ------
        pt : MatrixProductState
            The state at each of the times in ``ts``. This is a copy of
            internal state used, so inplace changes can be made to it.
        """
        # convert ts to list, to to calc range and use progress bar
        ts = sorted(ts)
        T = ts[-1]

        # need to use dt always so tol applies over whole T sweep
        dt = self._compute_sweep_dt_tol(T, dt, tol, order)

        # set up progress bar
        progbar = self.progbar if (progbar is None) else progbar
        if progbar:
            ts = qu.utils.progbar(ts)

        for t in ts:
            self.update_to(t, dt=dt, tol=False, order=order, progbar=False)

            if progbar:
                self._set_progbar_desc(ts)

            yield self.pt


def OTOC_local(psi0, H, H_back, ts, i, A, j=None, B=None,
               initial_eigenstate='check', **tebd_opts):
    """ The out-of-time-ordered correlator (OTOC) generating by two local
    operator A and B acting on site 'i', note it's a function of time.

    Parameters
    ----------
    psi0 : MatrixProductState
        The initial state in MPS form.
    H : NNI
        The Hamiltonian for forward time-evolution.
    H_back : NNI
        The Hamiltonian for backward time-evolution, should have only
        sign difference with 'H'.
    ts : sequence of float
        The time to evolve to.
    i : int
        The site where the local operators acting on.
    A : array
        The operator to act with.
    initial_eigenstate: {'check', Flase, True}
        To check the psi0 is or not eigenstate of operator B. If psi0 is the
        eigenstate of B, it will run a simpler version of OTOC calculation
        automatically.

    Returns
    ----------
    The OTOC <A(t)B(0)A(t)B(0)>
    """

    if B is None:
        B = A
    if j is None:
        j = i

    if initial_eigenstate == 'check':
        psi = psi0.gate(B, j, contract=True)
        x = psi0.H.expec(psi)
        y = psi.H.expec(psi)
        if abs(x**2 - y) < 1e-10:
            initial_eigenstate = True
        else:
            initial_eigenstate = False

    if initial_eigenstate is True:
        tebd1 = TEBD(psi0, H, **tebd_opts)
        x = psi0.H.expec(psi0.gate(B, j, contract=True))
        for t in ts:
            # evolve forward
            tebd1.update_to(t)
            # apply first A-gate
            psi_t_A = tebd1.pt.gate(A, i, contract=True)
            # evolve backwards
            tebd2 = TEBD(psi_t_A, H_back, **tebd_opts)
            tebd2.update_to(t)
            # compute expectation with second B-gate
            psi_f = tebd2.pt
            yield x * psi_f.H.expec(psi_f.gate(B, j, contract=True))
    else:
        # set the initial TEBD and apply the first operator A to right
        psi0_L = psi0
        tebd1_L = TEBD(psi0_L, H, **tebd_opts)

        psi0_R = psi0.gate(B, j, contract=True)
        tebd1_R = TEBD(psi0_R, H, **tebd_opts)

        for t in ts:
            # evolve forward
            tebd1_L.update_to(t)
            tebd1_R.update_to(t)

            # apply the opertor A to both left and right states
            psi_t_L_A = tebd1_L.pt.gate(A, i, contract=True)
            psi_t_R_A = tebd1_R.pt.gate(A.H, i, contract=True)

            # set the second left and right TEBD
            tebd2_L = TEBD(psi_t_L_A, H_back, **tebd_opts)
            tebd2_R = TEBD(psi_t_R_A, H_back, **tebd_opts)

            # evolve backwards
            tebd2_L.update_to(t)
            tebd2_R.update_to(t)

            # apply the laste operator B to left and compute overlap
            psi_f_L = tebd2_L.pt.gate(B.H, j, contract=True)
            psi_f_R = tebd2_R.pt
            yield psi_f_L.H.expec(psi_f_R)



"""Functions for more advanced calculations of quantities and properties of
quantum objects.
"""
import numbers
import itertools
import functools
import collections
from math import sin, cos, pi, log, log2, sqrt

import numpy as np
import numpy.linalg as nla
from scipy.optimize import minimize
from cytoolz import frequencies, keymap

from .core import (
    njit, issparse, isop, zeroify, realify, prod, isvec, dot, dag,
    qu, kron, eye, ikron, tr, ptr, infer_size, expec, dop, ensure_qarray,
)
from .linalg.base_linalg import (
    eigh, eigvalsh, norm, sqrtm, norm_trace_dense
)
from .linalg.approx_spectral import (
    entropy_subsys_approx, tr_sqrt_subsys_approx,
    logneg_subsys_approx, gen_bipartite_spectral_fn,
)
from .gen.operators import pauli
from .gen.states import (
    basis_vec, bell_state, bloch_state
)
from .utils import int2tup, raise_cant_find_library_function

try:
    from opt_einsum import contract
except ImportError:
    contract = raise_cant_find_library_function('opt_einsum')


def fidelity(p1, p2, squared=False):
    """Fidelity between two quantum states. By default, the unsquared fidelity
    is used, equivalent in the pure state case to ``|<psi|phi>|``.

    Parameters
    ----------
    p1 : vector or operator
        First state.
    p2 : vector or operator
        Second state.
    squared : bool, optional
        Whether to use the squared convention or not, by default False.

    Returns
    -------
    float
    """
    if isvec(p1) or isvec(p2):
        F = expec(p1, p2).real
        if not squared:
            F = max(0.0, F)**0.5
        return F

    sqrho = sqrtm(p1)
    F = tr(sqrtm(dot(sqrho, dot(p2, sqrho)))).real

    if squared:
        F = F**2

    return F


def purify(rho):
    """Take state rho and purify it into a wavefunction of squared
    dimension.

    Parameters
    ----------
    rho : operator
        Density operator to purify.

    Returns
    -------
    vector :
        The purified ket.
    """
    d = rho.shape[0]
    evals, vs = eigh(rho)
    evals = np.sqrt(np.clip(evals, 0, 1))
    psi = np.zeros(shape=(d**2, 1), dtype=complex)
    for i, evals in enumerate(evals.flat):
        psi += evals * kron(vs[:, [i]], basis_vec(i, d))
    return qu(psi)


@ensure_qarray
def kraus_op(rho, Ek, dims=None, where=None, check=False):
    r"""Operate on a mixed state with a set of kraus operators:

    .. math::

        \sigma = \sum_k E_k \rho E_k^{\dagger}

    Parameters
    ----------
    rho : (d, d) density matrix
        The density operator to perform operation on.
    Ek : (K, d, d) array or sequence of K (d, d) arrays
        The Kraus operator(s).
    dims : sequence of int, optional
        The subdimensions of ``rho``.
    where : int or sequence of int, optional
        Which of the subsystems to apply the operation on.
    check : bool, optional
        Where to check ``sum_k(Ek.H @ Ek) == 1``.

    Returns
    -------
    sigma : density matrix
        The state after the kraus operation.

    Examples
    --------
    The depolarising channel:

    .. code-block:: python3

        In [1]: import quimb as qu

        In [2]: rho = qu.rand_rho(2)

        In [3]: I, X, Y, Z = (qu.pauli(s) for s in 'IXYZ')

        In [4]: [qu.expec(rho, A) for A in (X, Y, Z)]
        Out[4]: [-0.652176185230622, -0.1301762132792484, 0.22362918368272583]

        In [5]: p = 0.1

        In [6]: Ek = [
           ...:     (1 - p)**0.5 * I,
           ...:     (p / 3)**0.5 * X,
           ...:     (p / 3)**0.5 * Y,
           ...:     (p / 3)**0.5 * Z
           ...: ]

        In [7]: sigma = qu.kraus_op(rho, Ek)

        In [8]: [qu.expec(sigma, A) for A in (X, Y, Z)]
        Out[8]: [-0.5652193605332058, -0.11281938484201527, 0.1938119591916957]
    """
    if not isinstance(Ek, np.ndarray):
        Ek = np.stack(Ek, axis=0)

    if check:
        SEk = np.einsum('kij,kil', Ek.conj(), Ek)
        if norm(SEk - eye(Ek.shape[-1]), 'fro') > 1e-12:
            raise ValueError("Did not find ``sum(E_k.H @ Ek) == 1``.")

    if int(dims is None) + int(where is None) == 1:
        raise ValueError("If `dims` is specified so should `where`.")

    if isinstance(where, numbers.Integral):
        where = (where,)

    if dims:
        # the full subdimensions of ``rho``
        dims = tuple(dims)
        rho = rho.reshape(dims + dims)
        N = len(dims)

        # the subdimensions of where the kraus operator should act
        kdims = tuple(dims[i] for i in where)
        Ek = Ek.reshape((-1,) + kdims + kdims)

        rho_inds, out, Ei_inds, Ej_inds = [], [], ['K'], ['K']
        for i in range(N):
            if i in where:
                xi, xj = f'i{i}k', f'j{i}k'
                for inds in (rho_inds, Ei_inds):
                    inds.append(xi)
                for inds in (rho_inds, Ej_inds):
                    inds.append(xj)
                xi, xj = f'i{i}new', f'j{i}new'
                for inds in (out, Ei_inds):
                    inds.append(xi)
                for inds in (out, Ej_inds):
                    inds.append(xj)
            else:
                xi, xj = f'i{i}', f'j{i}'
                for inds in (rho_inds, out):
                    inds.append(xi)
                    inds.append(xj)
        for inds in (rho_inds, out, Ei_inds, Ej_inds):
            inds.sort()
    else:
        rho_inds = ['ik', 'jk']
        out = ['inew', 'jnew']
        Ei_inds = ['K', 'inew', 'ik']
        Ej_inds = ['K', 'jnew', 'jk']

    sigma = contract(Ek, Ei_inds, rho, rho_inds, Ek.conj(), Ej_inds, out)

    if dims:
        sigma = sigma.reshape(prod(dims), prod(dims))

    return sigma


@ensure_qarray
def projector(A, eigenvalue=1.0, tol=1e-12, autoblock=False):
    """Compute the projector for the target ``eigenvalue`` of operator
    ``A``.

    Parameters
    ----------
    A : operator or tuple[array, operator].
        The hermitian observable. If a tuple is supplied, assume this is the
        eigendecomposition of the observable.
    eigenvalue : float, optional
        The target eigenvalue to construct the projector for, default: 1.
    tol : float, optional
        The tolerance with which to select all eigenvalues.
    autoblock : bool, optional
        Whether to use automatic symmetry exploitation.

    Returns
    -------
    P : dense matrix
        The projector onto the target eigenspace.
    """
    if isinstance(A, (tuple, list)):
        el, ev = A
    else:
        el, ev = eigh(A, autoblock=autoblock)
    which = np.argwhere(abs(el - eigenvalue) < tol)
    P = np.zeros_like(ev)
    for i in which:
        vi = ev[:, i]
        P += (vi @ vi.H)
    return P


def measure(p, A, eigenvalue=None, tol=1e-12):
    r"""Measure state ``p`` with observable ``A``, collapsing the state in the
    process and returning the relevant eigenvalue.

    .. math::

        \left| \psi \right\rangle \rightarrow
        \frac{
            P_i \left| \psi \right\rangle
        }{
            \sqrt{\left\langle \psi \right| P_i \left| \psi \right\rangle}
        }

    and

    .. math::

        \rho \rightarrow
        \frac{
            P_i \rho P_i^{\dagger}
        }{
            \text{Tr} \left[ P_i \rho \right]
        }

    along with the corresponding eigenvalue :math:`\lambda_i`.

    Parameters
    ----------
    p : vector or matrix
        The quantum state to measure.
    A : matrix or tuple[array, matrix]
        The hermitian operator corresponding to an observable. You can supply
        also a pre-diagonalized operator here as a tuple of eigenvalues and
        eigenvectors.
    tol : float, optional
        The tolerance within which to group eigenspaces.
    eigenvalue : float, optional
        If specified, deterministically collapse to this result. Otherwise
        randomly choose a result as in 'real-life'.

    Returns
    -------
    result : float
        The result of the measurement.
    p_after : vector or matrix
        The quantum state post measurement.
    """
    if isinstance(A, (tuple, list)):
        el, ev = A
    else:
        el, ev = eigh(A)

    js = np.arange(el.size)

    # compute prob of each eigenvector
    if isvec(p):
        pj = (abs(ev.H @ p)**2).flatten()
    else:
        pj = contract("jk,kl,lj->j", ev.H, p, ev).real

    # then choose one
    if eigenvalue is None:
        j = np.random.choice(js, p=pj)
        eigenvalue = el[j]

    # now combine whole eigenspace
    P = projector((el, ev), eigenvalue=eigenvalue, tol=tol)
    total_prob = np.sum(pj[abs(el - eigenvalue) < tol])

    # now collapse the state
    if isvec(p):
        p_after = P @ (p / total_prob**0.5)
    else:
        p_after = (P @ p @ P.H) / total_prob

    return eigenvalue, p_after


def simulate_counts(p, C, phys_dim=2, seed=None):
    """Simulate measuring each qubit of ``p`` in the computational basis,
    producing output like that of ``qiskit``.

    Parameters
    ----------
    p : vector or operator
        The quantum state, assumed to be normalized, as either a ket or density
        operator.
    C : int
        The number of counts to perform.
    phys_dim : int, optional
        The assumed size of the subsystems of ``p``, defaults to 2 for qubits.

    Returns
    -------
    results : dict[str, int]
        The counts for each bit string measured.

    Examples
    --------

    Simulate measuring the state of each qubit in a GHZ-state:

    .. code:: python3

        >>> import quimb as qu
        >>> psi = qu.ghz_state(3)
        >>> qu.simulate_counts(psi, 1024)
        {'000': 514, '111': 510}

    """
    if seed is not None:
        np.random.seed(seed)

    n = infer_size(p, phys_dim)
    d = phys_dim**n

    if isop(p):
        pi = np.diag(p).real
    else:
        pi = np.multiply(np.conj(p), p).real

    # probability of each basis state
    pi = pi.reshape(-1)

    # raw counts in terms of integers
    raw_counts = np.random.choice(np.arange(d), size=C, p=pi)

    # convert to frequencies of binary
    bin_str = '{:0>' + str(n) + 'b}'
    results = keymap(bin_str.format, frequencies(raw_counts))

    return results


def dephase(rho, p, rand_rank=None):
    """Dephase ``rho`` by amount ``p``, that is, mix it
    with the maximally mixed state:

        rho -> (1 - p) * rho + p * I / d

    Parameters
    ----------
    rho : operator
        The state.
    p : float
        The final proportion of identity.
    rand_rank : int or float, optional
        If given, dephase with a random diagonal operator with this many
        non-zero entries. If float, proportion of full size.

    Returns
    -------
    rho_dephase : operator
        The dephased density operator.
    """
    d = rho.shape[0]

    if (rand_rank is None) or (rand_rank == d) or (rand_rank == 1.0):
        dephaser = eye(d) / d

    else:
        if not isinstance(rand_rank, numbers.Integral):
            rand_rank = int(rand_rank * d)
        rand_rank = min(max(1, rand_rank), d)

        dephaser = np.zeros((d, d))
        dephaser_diag = np.einsum("aa->a", dephaser)
        nnz = np.random.choice(np.arange(d), size=rand_rank, replace=False)
        dephaser_diag[nnz] = 1 / rand_rank

    return (1 - p) * rho + p * dephaser


@zeroify
def entropy(a, rank=None):
    """Compute the (von Neumann) entropy.

    Parameters
    ----------
    a : operator or 1d array
        Positive operator or list of positive eigenvalues.
    rank : int (optional)
        If operator has known rank, then a partial decomposition can be
        used to accelerate the calculation.

    Returns
    -------
    float
        The von Neumann entropy.

    See Also
    --------
    mutinf, entropy_subsys, entropy_subsys_approx
    """
    a = np.asarray(a)
    if np.ndim(a) == 1:
        evals = a
    else:
        if rank is None:
            evals = eigvalsh(a)
        else:  # know that not all eigenvalues needed
            evals = eigvalsh(a, k=rank, which='LM', backend='AUTO')

    evals = evals[evals > 0.0]
    return np.sum(-evals * np.log2(evals))


entropy_subsys = gen_bipartite_spectral_fn(entropy, entropy_subsys_approx, 0.0)
"""Calculate the entropy of a pure states' subsystem, optionally switching
to an approximate lanczos method when the subsystem is very large.

Parameters
----------
psi_ab : vector
    Bipartite state.
dims : sequence of int
    The sub-dimensions of the state.
sysa :  sequence of int
    The indices of which dimensions to calculate the entropy for.
approx_thresh : int, optional
    The size of sysa at which to switch to the approx method. Set to
    ``None`` to never use the approximation.
**approx_opts
    Supplied to :func:`entropy_subsys_approx`, if used.

Returns
-------
float
    The subsytem entropy.

See Also
--------
entropy, entropy_subsys_approx, mutinf_subsys
"""


@zeroify
def mutinf(p, dims=(2, 2), sysa=0, rank=None):
    """Find the mutual information for a bipartition of a state.

    That is, ``H(A) + H(B) - H(AB)``, for von Neumann entropy ``H``, and two
    subsystems A and B.

    Parameters
    ----------
    p : vector or operator
        State, can be vector or operator.
    dims : tuple(int), optional
        Internal dimensions of state.
    sysa : int, optional
        Index of first subsystem, A.
    sysb : int, optional
        Index of second subsystem, B.
    rank : int, optional
        If known, the rank of rho_ab, to speed calculation of ``H(AB)`` up.
        For example, if ``p`` comes from tracing out three qubits from a
        system, then its rank is 2^3 = 8 etc.

    Returns
    -------
    float

    See Also
    --------
    entropy, mutinf_subsys, entropy_subsys_approx
    """
    sysa = int2tup(sysa)

    # mixed combined system
    if isop(p):
        # total
        hab = entropy(p, rank=rank)

        # subsystem a
        rhoa = ptr(p, dims, sysa)
        ha = entropy(rhoa)

        # need subsystem b as well
        sysb = tuple(i for i in range(len(dims)) if i not in sysa)
        rhob = ptr(p, dims, sysb)
        hb = entropy(rhob)

    # pure combined system
    else:
        hab = 0.0
        ha = hb = entropy_subsys(p, dims, sysa)

    return ha + hb - hab


mutual_information = mutinf


def check_dims_and_indices(dims, *syss):
    """Make sure all indices found in the tuples ``syss`` are in
    ``range(len(dims))``.
    """
    nsys = len(dims)
    all_sys = sum(syss, ())

    if not all(0 <= i < nsys for i in all_sys):
        raise ValueError(f"Indices specified in `sysa` and `sysb` must be "
                         "in range({nsys}) for dims {dims}.")


def mutinf_subsys(psi_abc, dims, sysa, sysb, approx_thresh=2**13,
                  **approx_opts):
    """Calculate the mutual information of two subsystems of a pure state,
    possibly using an approximate lanczos method for large subsytems.

    Parameters
    ----------
    psi_abc : vector
        Tri-partite pure state.
    dims : sequence of int
        The sub dimensions of the state.
    sysa : sequence of int
        The index(es) of the subsystem(s) to consider part of 'A'.
    sysb : sequence of int
        The index(es) of the subsystem(s) to consider part of 'B'.
    approx_thresh : int, optional
        The size of subsystem at which to switch to the approximate lanczos
        method. Set to ``None`` to never use the approximation.
    approx_opts
        Supplied to :func:`entropy_subsys_approx`, if used.

    Returns
    -------
    float
        The mutual information.

    See Also
    --------
    mutinf, entropy_subsys, entropy_subsys_approx, logneg_subsys
    """
    sysa, sysb = int2tup(sysa), int2tup(sysb)

    check_dims_and_indices(dims, sysa, sysb)

    sz_a = prod(d for i, d in enumerate(dims) if i in sysa)
    sz_b = prod(d for i, d in enumerate(dims) if i in sysb)
    sz_c = prod(dims) // (sz_a * sz_b)

    kws = {'approx_thresh': approx_thresh, **approx_opts}

    if sz_c == 1:
        hab = 0.0
        ha = hb = entropy_subsys(psi_abc, dims, sysa, **kws)
    else:
        hab = entropy_subsys(psi_abc, dims, sysa + sysb, **kws)
        ha = entropy_subsys(psi_abc, dims, sysa, **kws)
        hb = entropy_subsys(psi_abc, dims, sysb, **kws)

    return hb + ha - hab


def schmidt_gap(psi_ab, dims, sysa):
    """Find the schmidt gap of the bipartition of ``psi_ab``. That is, the
    difference between the two largest eigenvalues of the reduced density
    operator.

    Parameters
    ----------
    psi_ab : vector
        Bipartite state.
    dims : sequence of int
        The sub-dimensions of the state.
    sysa :  sequence of int
        The indices of which dimensions to calculate the entropy for.

    Returns
    -------
    float
    """
    sysa = int2tup(sysa)
    sz_a = prod(d for i, d in enumerate(dims) if i in sysa)
    sz_b = prod(dims) // sz_a

    # pure state
    if sz_b == 1:
        return 1.0

    # also check if system b is smaller, since spectrum is same for both
    if sz_b < sz_a:
        # if so swap things around
        sysb = [i for i in range(len(dims)) if i not in sysa]
        sysa = sysb

    rho_a = ptr(psi_ab, dims, sysa)
    el = eigvalsh(rho_a, k=2, which='LM')
    return abs(el[0] - el[1])


def tr_sqrt(A, rank=None):
    """Return the trace of the sqrt of a positive semidefinite operator.
    """
    if rank is None:
        el = eigvalsh(A, sort=False)
    else:
        el = eigvalsh(A, k=rank, which='LM', backend='AUTO')
    return np.sum(np.sqrt(el[el > 0.0]))


tr_sqrt_subsys = gen_bipartite_spectral_fn(tr_sqrt, tr_sqrt_subsys_approx, 1.0)
"""Compute the trace sqrt of a subsystem, possibly using an approximate
lanczos method when the subsytem is big.

Parameters
----------
psi_ab : vector
    Bipartite state.
dims : sequence of int
    The sub-dimensions of the state.
sysa :  sequence of int
    The indices of which dimensions to calculate the trace sqrt for.
approx_thresh : int, optional
    The size of sysa at which to switch to the approx method. Set to
    ``None`` to never use the approximation.
**approx_opts
    Supplied to :func:`tr_sqrt_subsys_approx`, if used.

Returns
-------
float
    The subsytem entropy.

See Also
--------
tr_sqrt, tr_sqrt_subsys_approx, partial_transpose_norm
"""


@ensure_qarray
def partial_transpose(p, dims=(2, 2), sysa=0):
    """Partial transpose of a density operator.

    Parameters
    ----------
    p : operator or vector
        The state to partially transpose.
    dims : tuple(int), optional
        The internal dimensions of the state.
    sysa : sequence of int
        The indices of 'system A', everything else assumed to be 'system B'.

    Returns
    -------
    operator

    See Also
    --------
    logneg, negativity
    """
    sysa = int2tup(sysa)

    ndims = len(dims)
    perm_ket_inds = []
    perm_bra_inds = []

    for i in range(ndims):
        if i in sysa:
            perm_ket_inds.append(i + ndims)
            perm_bra_inds.append(i)
        else:
            perm_ket_inds.append(i)
            perm_bra_inds.append(i + ndims)

    return (np.asarray(qu(p, "dop"))
            .reshape((*dims, *dims))
            .transpose((*perm_ket_inds, *perm_bra_inds))
            .reshape((prod(dims), prod(dims))))


def partial_transpose_norm(p, dims, sysa):
    """Compute the norm of the partial transpose for (log)-negativity,
    taking a shortcut (trace sqrt of reduced subsytem), when system is a
    vector.
    """
    sysa = int2tup(sysa)

    # check for pure bipartition -> easier to calc
    if isvec(p):
        sz_a = prod(d for i, d in enumerate(dims) if i in sysa)
        sz_b = prod(dims) // sz_a

        # check if system b is smaller, since entropy is same for both a & b.
        if sz_b < sz_a:
            # if so swap things around
            sysb = [i for i in range(len(dims)) if i not in sysa]
            sysa = sysb

        rhoa = ptr(p, dims, sysa)
        return tr_sqrt(rhoa)**2

    return norm_trace_dense(partial_transpose(p, dims, sysa), isherm=True)


@zeroify
def logneg(p, dims=(2, 2), sysa=0):
    """Compute logarithmic negativity between two subsytems.
    This is defined as  log_2( | rho_{AB}^{T_B} | ). This only handles
    bipartitions (including pure states efficiently), and will not trace
    anything out.

    Parameters
    ----------
    p : ket vector or density operator
        State to compute logarithmic negativity for.
    dims : tuple(int), optional
        The internal dimensions of ``p``.
    sysa : int, optional
        Index of the first subsystem, A, relative to ``dims``.

    Returns
    -------
    float

    See Also
    --------
    negativity, partial_transpose, logneg_subsys_approx
    """
    return max(0.0, log2(partial_transpose_norm(p, dims, sysa)))


logarithmic_negativity = logneg


def logneg_subsys(psi_abc, dims, sysa, sysb,
                  approx_thresh=2**13, **approx_opts):
    """Compute the logarithmic negativity between two subsystems of a pure
    state, possibly using an approximate lanczos for large subsystems. Uses
    a special method if the two subsystems form a bipartition of the state.

    Parameters
    ----------
    psi_abc : vector
        Tri-partite pure state.
    dims : sequence of int
        The sub dimensions of the state.
    sysa : sequence of int
        The index(es) of the subsystem(s) to consider part of 'A'.
    sysb : sequence of int
        The index(es) of the subsystem(s) to consider part of 'B'.
    approx_thresh : int, optional
        The size of subsystem at which to switch to the approximate lanczos
        method. Set to ``None`` to never use the approximation.
    approx_opts
        Supplied to :func:`~quimb.logneg_subsys_approx`, if used.

    Returns
    -------
    float
        The logarithmic negativity.

    See Also
    --------
    logneg, mutinf_subsys, logneg_subsys_approx
    """
    sysa, sysb = int2tup(sysa), int2tup(sysb)

    check_dims_and_indices(dims, sysa, sysb)

    sz_a = prod(d for i, d in enumerate(dims) if i in sysa)
    sz_b = prod(d for i, d in enumerate(dims) if i in sysb)
    sz_ab = sz_a * sz_b
    sz_c = prod(dims) // sz_ab

    # check for pure bipartition
    if sz_c == 1:
        psi_ab_ppt_norm = tr_sqrt_subsys(
            psi_abc, dims, sysa, approx_thresh=approx_thresh, **approx_opts)**2
        return max(log2(psi_ab_ppt_norm), 0.0)

    # check whether to use approx lanczos method
    if (approx_thresh is not None) and (sz_ab >= approx_thresh):
        return logneg_subsys_approx(psi_abc, dims, sysa, sysb, **approx_opts)

    rho_ab = ptr(psi_abc, dims, sysa + sysb)

    # need to adjust for new dimensions and indices
    new_dims, new_sysa = [], []
    new_inds = iter(range(len(dims)))

    for i, d in enumerate(dims):
        if i in sysa:
            new_dims.append(d)
            new_sysa.append(next(new_inds))
        elif i in sysb:
            new_dims.append(d)
            next(new_inds)  # don't need sysb

    return logneg(rho_ab, new_dims, new_sysa)


def negativity(p, dims=(2, 2), sysa=0):
    """Compute negativity between two subsytems.

    This is defined as  (| rho_{AB}^{T_B} | - 1) / 2. If ``len(dims) > 2``,
    then the non-target dimensions will be traced out first.

    Parameters
    ----------
    p : ket vector or density operator
        State to compute logarithmic negativity for.
    dims : tuple(int), optional
        The internal dimensions of ``p``.
    sysa : int, optional
        Index of the first subsystem, A, relative to ``dims``.

    Returns
    -------
    float

    See Also
    --------
    logneg, partial_transpose, negativity_subsys_approx
    """
    return max(0.0, (partial_transpose_norm(p, dims, sysa) - 1) / 2)


@zeroify
def concurrence(p, dims=(2, 2), sysa=0, sysb=1):
    """Concurrence of two-qubit state.

    If ``len(dims) > 2``, then the non-target dimensions will be traced out
    first.

    Parameters
    ----------
    p : ket vector or density operator
        State to compute concurrence for.
    dims : tuple(int), optional
        The internal dimensions of ``p``.
    sysa : int, optional
        Index of the first subsystem, A, relative to ``dims``.
    sysb : int, optional
        Index of the first subsystem, B, relative to ``dims``.

    Returns
    -------
    float
    """
    if len(dims) > 2:
        p = ptr(p, dims, (sysa, sysb))

    Y = pauli('Y')

    if isop(p):
        pt = dot(kron(Y, Y), dot(p.conj(), kron(Y, Y)))
        evals = (nla.eigvals(dot(p, pt)).real**2)**0.25
        return max(0, 2 * np.max(evals) - np.sum(evals))
    else:
        pt = dot(kron(Y, Y), p.conj())
        c = np.real(abs(dot(dag(p), pt))).item(0)
        return max(0, c)


def one_way_classical_information(p_ab, prjs, precomp_func=False):
    """One way classical information for two qubit density operator.

    Parameters
    ----------
    p_ab : operator
        State of two qubits
    prjs : sequence of matrices
       The POVMs.
    precomp_func : bool, optional
        Whether to return a pre-computed function, closed over the actual
        state.

    Returns
    -------
    float or callable
        The one-way classical information or the function to compute it for
        the given state which takes a set of POVMs as its single argument.
    """
    p_a = ptr(p_ab, (2, 2), 0)
    s_a = entropy(p_a)

    def owci(prjs):
        def gen_paj():
            for prj in prjs:
                p_ab_j = dot((eye(2) & prj), p_ab)
                prob = tr(p_ab_j)
                p_a_j = ptr(p_ab_j, (2, 2), 0) / prob
                yield prob, p_a_j
        return s_a - sum(p * entropy(rho) for p, rho in gen_paj())

    return owci if precomp_func else owci(prjs)


@zeroify
def quantum_discord(p, dims=(2, 2), sysa=0, sysb=1):
    """Quantum Discord for two qubit density operator.

    If ``len(dims) > 2``, then the non-target dimensions will be traced out
    first.

    Parameters
    ----------
    p : ket vector or density operator
        State to compute quantum discord for.
    dims : tuple(int), optional
        The internal dimensions of ``p``.
    sysa : int, optional
        Index of the first subsystem, A, relative to ``dims``.
    sysb : int, optional
        Index of the first subsystem, B, relative to ``dims``.

    Returns
    -------
    float
    """
    if len(dims) > 2:
        p = ptr(p, dims, (sysa, sysb))
    else:
        p = qu(p, "dop")
    iab = mutual_information(p)
    owci = one_way_classical_information(p, None, precomp_func=True)

    def trial_qd(a):
        ax, ay, az = sin(a[0]) * cos(a[1]), sin(a[0]) * sin(a[1]), cos(a[0])
        prja = bloch_state(ax, ay, az)
        prjb = eye(2) - prja
        return iab - owci((prja, prjb))

    opt = minimize(trial_qd, (pi / 2, pi),
                   method="SLSQP", bounds=((0, pi), (0, 2 * pi)))
    if opt.success:
        return opt.fun
    else:  # pragma: no cover
        raise ValueError(opt.message)


@zeroify
def trace_distance(p1, p2):
    """Trace distance between two states.

    Parameters
    ----------
    p1 : ket or density operator
        The first state.
    p2 : ket or density operator
        The second state.

    Returns
    -------
    float
    """
    p1_is_op, p2_is_op = isop(p1), isop(p2)

    # If both are pure kets then special case
    if (not p1_is_op) and (not p2_is_op):
        return sqrt(1 - expec(p1, p2))

    # Otherwise do full calculation
    return 0.5 * norm((p1 if p1_is_op else dop(p1)) -
                      (p2 if p2_is_op else dop(p2)), "tr")


def cprint(psi, prec=6):
    """Print a state in the computational basis.

    Parameters
    ----------
    psi : vector
        The pure state.
    prec : int, optional
        How many significant figures to print.

    Examples
    --------

        >>> cprint(rand_ket(2**2))
        (-0.0751069-0.192635j) |00>
          (0.156837-0.562213j) |01>
        (-0.307381+0.0291168j) |10>
          (-0.14302+0.707661j) |11>

        >>> cprint(w_state(4))
        (0.5+0j) |0001>
        (0.5+0j) |0010>
        (0.5+0j) |0100>
        (0.5+0j) |1000>
    """
    d = psi.shape[0]
    n = int(log2(d))
    if 2**n != d:
        raise ValueError("State is not factorizable into computational basis.")

    coeff_str = "{:." + str(prec) + "}"

    coeffs, cs = [], []
    for c in itertools.product('01', repeat=n):
        c = "".join(c)
        ix = int(c, 2)
        coeff = psi.item(ix)
        if coeff != 0.0:
            cs.append(c)
            coeffs.append(coeff_str.format(coeff))

    max_str_len = max(map(len, coeffs))
    full_str = "{:>" + str(max_str_len) + "} |{}>"
    for coeff, c in zip(coeffs, cs):
        print(full_str.format(coeff, c))


def decomp(a, fn, fn_args, fn_d, nmlz_func, mode="p", tol=1e-3):
    """Decomposes an operator via the Hilbert-schmidt inner product.

    Can both print the decomposition or return it.

    Parameters
    ----------
    a : ket or density operator
        Operator to decompose.
    fn : callable
        Function to generate operator/state to decompose with.
    fn_args :
        Sequence of args whose permutations will be supplied to ``fn``.
    fn_d : int
        The dimension of the operators that `fn` produces.
    nmlz_func : callable
        Function to produce a normlization coefficient given the ``n``
        permutations of operators that will be produced.
    mode :
        String, include ``'p'`` to print the decomp and/or ``'c'`` to
        return OrderedDict, sorted by size of contribution.
    tol :
        Print operators with contirbution above ``tol`` only.

    Returns
    -------
    None or OrderedDict:
        Pauli operator name and expec with ``a``.

    See Also
    --------
    pauli_decomp, bell_decomp
    """
    if isvec(a):
        a = qu(a, "dop")  # make sure operator
    n = infer_size(a, base=fn_d)

    # define generator for inner product to iterate over efficiently
    def calc_name_and_overlap():
        for perm in itertools.product(fn_args, repeat=n):
            op = kron(*(fn(x, sparse=True) for x in perm)) * nmlz_func(n)
            cff = expec(a, op)
            yield "".join(str(x) for x in perm), cff

    names_cffs = list(calc_name_and_overlap())
    # sort by descending expec and turn into OrderedDict
    names_cffs.sort(key=lambda pair: -abs(pair[1]))
    names_cffs = collections.OrderedDict(names_cffs)
    # Print decomposition
    if "p" in mode:
        for x, cff in names_cffs.items():
            if abs(cff) < 0.01:
                break
            dps = int(round(0.5 - np.log10(1.001 * tol)))  # decimal places
            print(x, "{: .{prec}f}".format(cff, prec=dps))
    # Return full calculation
    if "c" in mode:
        return names_cffs


pauli_decomp = functools.partial(decomp,
                                 fn=pauli,
                                 fn_args='IXYZ',
                                 fn_d=2,
                                 nmlz_func=lambda n: 2**-n)
"""Decompose an operator into Paulis."""

bell_decomp = functools.partial(decomp,
                                fn=bell_state,
                                fn_args=(0, 1, 2, 3),
                                fn_d=4,
                                nmlz_func=lambda x: 1)
"""Decompose an operator into bell-states."""


def correlation(p, A, B, sysa, sysb, dims=None, sparse=None,
                precomp_func=False):
    """Calculate the correlation between two sites given two operators.

    Parameters
    ----------
    p : ket or density operator
        State to compute correlations for, ignored if ``precomp_func=True``.
    A : operator
        Operator to act on first subsystem.
    B : operator
        Operator to act on second subsystem.
    sysa : int
        Index of first subsystem.
    sysb : int
        Index of second subsystem.
    dims : tuple of int, optional
        Internal dimensions of ``p``, will be assumed to be qubits if not
        given.
    sparse : bool, optional
        Whether to compute with sparse operators.
    precomp_func : bool, optional
        Whether to return result or single arg function closed
        over precomputed operator.

    Returns
    -------
    float or callable
        The correlation, <ab> - <a><b>, or a function to compute for an
        arbitrary state.
    """
    if dims is None:
        sz_p = infer_size(p)
        dims = (2,) * sz_p
    if sparse is None:
        sparse = issparse(A) or issparse(B)

    opts = {'sparse': sparse,
            'coo_build': sparse,
            'stype': 'csr' if sparse else None}
    opab = ikron((A, B), dims, (sysa, sysb), **opts)
    A = ikron((A,), dims, sysa, **opts)
    B = ikron((B,), dims, sysb, **opts)

    @realify
    def corr(state):
        return expec(opab, state) - expec(A, state) * expec(B, state)

    return corr if precomp_func else corr(p)


def pauli_correlations(p, ss=("xx", "yy", "zz"), sysa=0, sysb=1,
                       sum_abs=False, precomp_func=False):
    """Calculate the correlation between sites for a list of operator pairs
    choisen from the pauli matrices.

    Parameters
    ----------
    p : ket or density operator
        State to compute correlations for. Ignored if ``precomp_func=True``.
    ss : tuple or str
        List of pairs specifiying pauli matrices.
    sysa : int, optional
        Index of first site.
    sysb : int, optional
        Index of second site.
    sum_abs : bool, optional
        Whether to sum over the absolute values of each correlation
    precomp_func : bool, optional
        whether to return the values or a single argument
        function closed over precomputed operators etc.

    Returns
    -------
    list of float, list of callable, float or callable
        Either the value(s) of each correlation or the function(s) to compute
        the correlations for an arbitrary state, depending on ``sum_abs`` and
        ``precomp_func``.
    """
    def gen_corr_list():
        for s1, s2 in ss:
            yield correlation(p, pauli(s1), pauli(s2), sysa, sysb,
                              precomp_func=precomp_func)

    if sum_abs:

        if precomp_func:
            return lambda p: sum((abs(corr(p)) for corr in gen_corr_list()))

        return sum((abs(corr) for corr in gen_corr_list()))

    return tuple(gen_corr_list())


def ent_cross_matrix(p, sz_blc=1, ent_fn=logneg, calc_self_ent=True,
                     upscale=False):
    """Calculate the pair-wise function ent_fn  between all sites or blocks
    of a state.

    Parameters
    ----------
    p : ket or density operator
        State.
    sz_blc : int
        Size of the blocks to partition the state into. If the number of
        individual sites is not a multiple of this then the final (smaller)
        block will be ignored.
    ent_fn : callable
        Bipartite function, notionally entanglement
    calc_self_ent : bool
        Whether to calculate the function for each site
        alone, purified. If the whole state is pure then this is the
        entanglement with the whole remaining system.
    upscale : bool, optional
        Whether, if sz_blc != 1, to upscale the results so that the output
        array is the same size as if it was.

    Returns
    -------
    2D-array
        array of pairwise ent_fn results.
    """

    sz_p = infer_size(p)
    dims = (2,) * sz_p
    n = sz_p // sz_blc
    ents = np.empty((n, n))

    ispure = isvec(p)
    if ispure and sz_blc * 2 == sz_p:  # pure bipartition
        ent = ent_fn(p, dims=(2**sz_blc, 2**sz_blc)) / sz_blc
        ents[:, :] = ent
        if not calc_self_ent:
            for i in range(n):
                ents[i, i] = np.nan

    else:
        # Range over pairwise blocks
        for i in range(0, sz_p - sz_blc + 1, sz_blc):
            for j in range(i, sz_p - sz_blc + 1, sz_blc):
                if i == j:
                    if calc_self_ent:
                        rhoa = ptr(p, dims, [i + b for b in range(sz_blc)])
                        psiap = purify(rhoa)
                        ent = ent_fn(psiap,
                                     dims=(2**sz_blc, 2**sz_blc)) / sz_blc
                    else:
                        ent = np.nan
                else:
                    rhoab = ptr(p, dims, [i + b for b in range(sz_blc)] +
                                         [j + b for b in range(sz_blc)])
                    ent = ent_fn(rhoab, dims=(2**sz_blc, 2**sz_blc)) / sz_blc
                ents[i // sz_blc, j // sz_blc] = ent
                ents[j // sz_blc, i // sz_blc] = ent

    if upscale:
        up_ents = np.tile(np.nan, (sz_p, sz_p))

        for i in range(n):
            for j in range(i, n):
                up_ents[i * sz_blc:(i + 1) * sz_blc,
                        j * sz_blc:(j + 1) * sz_blc] = ents[i, j]
                up_ents[j * sz_blc:(j + 1) * sz_blc,
                        i * sz_blc:(i + 1) * sz_blc] = ents[j, i]

        ents = up_ents

    return ents


def qid(p, dims, inds, precomp_func=False, sparse_comp=True,
        norm_func=norm, power=2, coeff=1):
    # Check inputs
    inds = (inds,) if isinstance(inds, numbers.Number) else inds

    # Construct operators
    ops_i = tuple(tuple(ikron(pauli(s), dims, ind, sparse=sparse_comp)
                        for s in "xyz")
                  for ind in inds)

    # Define function closed over precomputed operators
    def qid_func(x):
        if isvec(x):
            x = dop(x)
        return tuple(sum(coeff * norm_func(dot(x, op) - dot(op, x))**power
                         for op in ops)
                     for ops in ops_i)

    return qid_func if precomp_func else qid_func(p)


def is_degenerate(op, tol=1e-12):
    """Check if operator has any degenerate eigenvalues, determined relative
    to mean spacing of all eigenvalues.

    Parameters
    ----------
    op : operator or 1d-array
        Operator or assumed eigenvalues to check degeneracy for.
    tol : float
        How much closer than evenly spaced the eigenvalue gap has to be
        to count as degenerate.

    Returns
    -------
    n_dgen : int
        Number of degenerate eigenvalues.
    """
    op = np.asarray(op)
    if op.ndim != 1:
        evals = eigvalsh(op)
    else:
        evals = op
    l_gaps = evals[1:] - evals[:-1]
    l_tol = tol * (evals[-1] - evals[0]) / op.shape[0]
    return np.count_nonzero(abs(l_gaps) < l_tol)


def is_eigenvector(x, A, tol=1e-14):
    """Determines whether a vector is an eigenvector of an operator.

    Parameters
    ----------
    x : vector
        Vector to check.
    A : operator
        Matrix to check.
    tol : float, optional
        The variance must be smaller than this value.

    Returns
    -------
    bool
        Whether ``A @ x = l * x`` for some scalar ``l``.
    """
    mat_vec = dot(A, x)
    E = expec(x, mat_vec)
    E2 = expec(x, dot(A, mat_vec))
    return abs(E**2 - E2) < tol


@njit
def page_entropy(sz_subsys, sz_total):  # pragma: no cover
    """Calculate the page entropy, i.e. expected entropy for a subsytem
    of a random state in Hilbert space.

    Parameters
    ----------
    sz_subsys : int
        Dimension of subsystem.
    sz_total : int
        Dimension of total system.

    Returns
    -------
    s : float
        Entropy in bits.
    """
    if sz_subsys > sz_total**0.5:
        sz_subsys = sz_total // sz_subsys

    n = sz_total // sz_subsys

    s = 0
    for k in range(n + 1, sz_total + 1):
        s += 1 / k
    s -= (sz_subsys - 1) / (2 * n)

    # Normalize into bits of entropy
    return s / log(2)


def heisenberg_energy(L):
    """Get the analytic isotropic heisenberg chain ground energy for length L.
    Useful for testing. Assumes the heisenberg model is defined with spin
    operators not pauli matrices (overall factor of 2 smaller). Taken from [1].

    [1] Nickel, Bernie. "Scaling corrections to the ground state energy
    of the spin-½ isotropic anti-ferromagnetic Heisenberg chain." Journal of
    Physics Communications 1.5 (2017): 055021

    Parameters
    ----------
    L : int
        The length of the chain.

    Returns
    -------
    energy : float
        The ground state energy.
    """
    Einf = (0.5 - 2 * log(2)) * L
    Efinite = pi**2 / (6 * L)
    correction = 1 + 0.375 / log(L)**3
    return (Einf - Efinite * correction) / 2



import numbers

import cytoolz
from autoray import do

import quimb as qu
from .tensor_core import get_tags, PTensor
from .tensor_gen import MPS_computational_state
from .tensor_1d import TensorNetwork1DVector
from . import array_ops as ops


def _convert_ints_and_floats(x):
    if isinstance(x, str):
        try:
            return int(x)
        except ValueError:
            pass

        try:
            return float(x)
        except ValueError:
            pass

    return x


def _put_registers_last(x):
    # no need to do anything unless parameter (i.e. float) is found last
    if not isinstance(x[-1], float):
        return x

    # swap this last group of floats with the penultimate group of integers
    parts = tuple(cytoolz.partitionby(type, x))
    return tuple(cytoolz.concatv(*parts[:-2], parts[-1], parts[-2]))


def parse_qasm(qasm):
    """Parse qasm from a string.

    Parameters
    ----------
    qasm : str
        The full string of the qasm file.

    Returns
    -------
    circuit_info : dict
        Information about the circuit:

        - circuit_info['n']: the number of qubits
        - circuit_info['n_gates']: the number of gates in total
        - circuit_info['gates']: list[list[str]], list of gates, each of which
          is a list of strings read from a line of the qasm file.
    """

    lines = qasm.split('\n')
    n = int(lines[0])

    # turn into tuples of python types
    gates = [
        tuple(map(_convert_ints_and_floats, line.strip().split(" ")))
        for line in lines[1:] if line
    ]

    # put registers/parameters in standard order and detect if gate round used
    gates = tuple(map(_put_registers_last, gates))
    round_specified = isinstance(gates[0][0], numbers.Integral)

    return {
        'n': n,
        'gates': gates,
        'n_gates': len(gates),
        'round_specified': round_specified,
    }


def parse_qasm_file(fname, **kwargs):
    """Parse a qasm file.
    """
    return parse_qasm(open(fname).read(), **kwargs)


def parse_qasm_url(url, **kwargs):
    """Parse a qasm url.
    """
    from urllib import request
    return parse_qasm(request.urlopen(url).read().decode(), **kwargs)


# -------------------------- core gate functions ---------------------------- #

def _merge_tags(tags, gate_opts):
    if isinstance(tags, str):
        tags = {tags}
    else:
        tags = set(tags)
    return tags | set(gate_opts.pop('tags', ()))


def build_gate_1(gate, tags=None):
    """Build a function that applies ``gate`` to a tensor network wavefunction.
    """

    def apply_constant_single_qubit_gate(psi, i, **gate_opts):
        mtags = _merge_tags(tags, gate_opts)
        psi.gate_(gate, int(i), tags=mtags, **gate_opts)

    return apply_constant_single_qubit_gate


def build_gate_2(gate, tags=None):
    """Build a function that applies ``gate`` to a tensor network wavefunction.
    """

    def apply_constant_two_qubit_gate(psi, i, j, **gate_opts):
        mtags = _merge_tags(tags, gate_opts)
        psi.gate_(gate, (int(i), int(j)), tags=mtags, **gate_opts)

    return apply_constant_two_qubit_gate


# non tensor gates

def apply_swap(psi, i, j, **gate_opts):
    iind, jind = map(psi.site_ind, (int(i), int(j)))
    psi.reindex_({iind: jind, jind: iind})


# parametrizable gates

def rx_gate_param_gen(params):
    phi, = params
    c = do('cos', phi / 2)
    s = -1j * do('sin', phi / 2)
    data = [[c, s], [s, c]]
    return do('array', data, like=params)


def apply_Rx(psi, theta, i, parametrize=False, **gate_opts):
    """Apply an X-rotation of ``theta`` to tensor network wavefunction ``psi``.
    """
    mtags = _merge_tags({'RX'}, gate_opts)
    if parametrize:
        G = ops.PArray(rx_gate_param_gen, (float(theta),))
    else:
        G = qu.Rx(float(theta))
    psi.gate_(G, int(i), tags=mtags, **gate_opts)


def ry_gate_param_gen(params):
    phi, = params
    c = do('cos', phi / 2)
    s = do('sin', phi / 2)
    data = [[c, -s], [s, c]]
    return do('array', data, like=params)


def apply_Ry(psi, theta, i, parametrize=False, **gate_opts):
    """Apply a Y-rotation of ``theta`` to tensor network wavefunction ``psi``.
    """
    mtags = _merge_tags({'RY'}, gate_opts)
    if parametrize:
        G = ops.PArray(ry_gate_param_gen, (float(theta),))
    else:
        G = qu.Ry(float(theta))
    psi.gate_(G, int(i), tags=mtags, **gate_opts)


def rz_gate_param_gen(params):
    phi, = params
    c = do('cos', phi / 2)
    s = -1j * do('sin', phi / 2)
    data = [[c + s, 0], [0, c - s]]
    return do('array', data, like=params)


def apply_Rz(psi, theta, i, parametrize=False, **gate_opts):
    """Apply a Z-rotation of ``theta`` to tensor network wavefunction ``psi``.
    """
    mtags = _merge_tags({'RZ'}, gate_opts)
    if parametrize:
        G = ops.PArray(rz_gate_param_gen, (float(theta),))
    else:
        G = qu.Rz(float(theta))
    psi.gate_(G, int(i), tags=mtags, **gate_opts)


def u3_gate_param_gen(params):
    theta, phi, lamda = params

    c2 = do('cos', theta / 2)
    s2 = do('sin', theta / 2)
    el = do('exp', 1.j * lamda)
    ep = do('exp', 1.j * phi)
    elp = do('exp', 1.j * (lamda + phi))

    data = [[c2, -el * s2],
            [ep * s2, elp * c2]]

    return do('array', data, like=params)


def apply_U3(psi, theta, phi, lamda, i, parametrize=False, **gate_opts):
    mtags = _merge_tags({'U3'}, gate_opts)
    if parametrize:
        G = ops.PArray(u3_gate_param_gen, (theta, phi, lamda))
    else:
        G = qu.U_gate(theta, phi, lamda)
    psi.gate_(G, int(i), tags=mtags, **gate_opts)


def fsim_param_gen(params):
    theta, phi = params

    a = do('cos', theta)
    b = -1j * do('sin', theta)
    c = do('exp', -1j * phi)

    data = [[[[1, 0],
              [0, 0]],
             [[0, a],
              [b, 0]]],
            [[[0, b],
              [a, 0]],
             [[0, 0],
              [0, c]]]]

    return do('array', data, like=params)


def apply_fsim(psi, theta, phi, i, j, parametrize=False, **gate_opts):
    mtags = _merge_tags({'FSIM'}, gate_opts)
    if parametrize:
        G = ops.PArray(fsim_param_gen, (theta, phi))
    else:
        G = qu.fsim(theta, phi)
    psi.gate_(G, (int(i), int(j)), tags=mtags, **gate_opts)


GATE_FUNCTIONS = {
    # constant single qubit gates
    'H': build_gate_1(qu.hadamard(), tags='H'),
    'X': build_gate_1(qu.pauli('X'), tags='X'),
    'Y': build_gate_1(qu.pauli('Y'), tags='Y'),
    'Z': build_gate_1(qu.pauli('Z'), tags='Z'),
    'S': build_gate_1(qu.S_gate(), tags='S'),
    'T': build_gate_1(qu.T_gate(), tags='T'),
    'X_1_2': build_gate_1(qu.Xsqrt(), tags='X_1/2'),
    'Y_1_2': build_gate_1(qu.Ysqrt(), tags='Y_1/2'),
    'Z_1_2': build_gate_1(qu.Zsqrt(), tags='Z_1/2'),
    'W_1_2': build_gate_1(qu.Wsqrt(), tags='W_1/2'),
    'HZ_1_2': build_gate_1(qu.Wsqrt(), tags='W_1/2'),
    # constant two qubit gates
    'CNOT': build_gate_2(qu.CNOT(), tags='CNOT'),
    'CX': build_gate_2(qu.cX(), tags='CX'),
    'CY': build_gate_2(qu.cY(), tags='CY'),
    'CZ': build_gate_2(qu.cZ(), tags='CZ'),
    'IS': build_gate_2(qu.iswap(), tags='ISWAP'),
    'ISWAP': build_gate_2(qu.iswap(), tags='ISWAP'),
    # special non-tensor gates
    'IDEN': lambda *args, **kwargs: None,
    'SWAP': apply_swap,
    # parametrizable gates
    'RX': apply_Rx,
    'RY': apply_Ry,
    'RZ': apply_Rz,
    'U3': apply_U3,
    'FS': apply_fsim,
    'FSIM': apply_fsim,
}

ONE_QUBIT_PARAM_GATES = {'RX', 'RY', 'RZ', 'U3'}
TWO_QUBIT_PARAM_GATES = {'FS', 'FSIM'}
ALL_PARAM_GATES = ONE_QUBIT_PARAM_GATES | TWO_QUBIT_PARAM_GATES


# --------------------------- main circuit class ---------------------------- #

class Circuit:
    """Class for simulating quantum circuits using tensor networks.

    Parameters
    ----------
    N : int, optional
        The number of qubits.
    psi0 : TensorNetwork1DVector, optional
        The initial state, assumed to be ``|00000....0>`` if not given.
    gate_opts : dict_like, optional
        Default keyword arguments to supply to each
        :func:`~quimb.tensor.tensor_1d.gate_TN_1D` call during the circuit.
    tags : str or sequence of str, optional
        Tag(s) to add to the initial wavefunction tensors (whether these are
        propagated to the rest of the circuit's tensors depends on
        ``gate_opts``).

    Attributes
    ----------
    psi : TensorNetwork1DVector
        The current wavefunction.

    Examples
    --------

    Create 3-qubit GHZ-state:

        >>> qc = qtn.Circuit(3)
        >>> gates = [
                ('H', 0),
                ('H', 1),
                ('CNOT', 1, 2),
                ('CNOT', 0, 2),
                ('H', 0),
                ('H', 1),
                ('H', 2),
            ]
        >>> qc.apply_gates(gates)
        >>> qc.psi
        <TensorNetwork1DVector(tensors=12, indices=14, nsites=3)>

        >>> qc.psi.to_dense().round(4)
        qarray([[ 0.7071+0.j],
                [ 0.    +0.j],
                [ 0.    +0.j],
                [-0.    +0.j],
                [-0.    +0.j],
                [ 0.    +0.j],
                [ 0.    +0.j],
                [ 0.7071+0.j]])
    """

    def __init__(self, N=None, psi0=None, gate_opts=None, tags=None):

        if N is None and psi0 is None:
            raise ValueError("You must supply one of `N` or `psi0`.")

        elif psi0 is None:
            self.N = N
            self._psi = MPS_computational_state('0' * N)

        elif N is None:
            self._psi = psi0.copy()
            self.N = psi0.nsites

        else:
            if N != psi0.nsites:
                raise ValueError("`N` doesn't match `psi0`.")
            self.N = N
            self._psi = psi0.copy()

        if tags is not None:
            if isinstance(tags, str):
                tags = (tags,)
            for tag in tags:
                self._psi.add_tag(tag)

        self.gate_opts = {} if gate_opts is None else dict(gate_opts)
        self.gate_opts.setdefault('contract', 'auto-split-gate')
        self.gate_opts.setdefault('propagate_tags', 'register')
        self.gates = []

        # when we add gates we will modify the TN structure, apart from in the
        # 'swap+split' case, which explicitly maintains an MPS
        if self.gate_opts['contract'] != 'swap+split':
            self._psi.view_as_(TensorNetwork1DVector)

    @classmethod
    def from_qasm(cls, qasm, **quantum_circuit_opts):
        """Generate a ``Circuit`` instance from a qasm string.
        """
        info = parse_qasm(qasm)
        qc = cls(info['n'], **quantum_circuit_opts)
        qc.apply_gates(info['gates'])
        return qc

    @classmethod
    def from_qasm_file(cls, fname, **quantum_circuit_opts):
        """Generate a ``Circuit`` instance from a qasm file.
        """
        info = parse_qasm_file(fname)
        qc = cls(info['n'], **quantum_circuit_opts)
        qc.apply_gates(info['gates'])
        return qc

    @classmethod
    def from_qasm_url(cls, url, **quantum_circuit_opts):
        """Generate a ``Circuit`` instance from a qasm url.
        """
        info = parse_qasm_url(url)
        qc = cls(info['n'], **quantum_circuit_opts)
        qc.apply_gates(info['gates'])
        return qc

    def apply_gate(self, gate_id, *gate_args, gate_round=None, **gate_opts):
        """Apply a single gate to this tensor network quantum circuit. If
        ``gate_round`` is supplied the tensor(s) added will be tagged with
        ``'ROUND_{gate_round}'``. Alternatively, putting an integer first like
        so::

            circuit.apply_gate(10, 'H', 7)

        Is automatically translated to::

            circuit.apply_gate('H', 7, gate_round=10)

        Parameters
        ----------
        gate_id : str
            Which type of gate to apply.
        gate_args : list[str]
            The argument to supply to it.
        gate_round : int, optional
            The gate round. If ``gate_id`` is integer-like, will also be taken
            from here, with then ``gate_id, gate_args = gate_args[0],
            gate_args[1:]``.
        gate_opts
            Supplied to the gate function, options here will override the
            default ``gate_opts``.
        """

        # unique tag
        tags = {f'GATE_{len(self.gates)}'}

        # parse which 'round' of gates
        if (gate_round is not None):
            tags.add(f'ROUND_{gate_round}')
        elif isinstance(gate_id, numbers.Integral) or gate_id.isdigit():
            # gate round given as first entry of qasm line
            tags.add(f'ROUND_{gate_id}')
            gate_id, gate_args = gate_args[0], gate_args[1:]

        gate_id = gate_id.upper()
        gate_fn = GATE_FUNCTIONS[gate_id]

        # overide any default gate opts
        opts = {**self.gate_opts, **gate_opts}

        # handle parametrize kwarg for non-parametrizable gates
        if ('parametrize' in opts) and (gate_id not in ALL_PARAM_GATES):
            parametrize = opts.pop('parametrize')
            if parametrize:
                msg = f"The gate '{gate_id}' cannot be parametrized."
                raise ValueError(msg)
            # can pop+ignore if False

        # gate the TN!
        gate_fn(self._psi, *gate_args, tags=tags, **opts)

        # keep track of the gates applied
        self.gates.append((gate_id, *gate_args))

    def apply_gates(self, gates):
        """Apply a sequence of gates to this tensor network quantum circuit.

        Parameters
        ----------
        gates : list[list[str]]
            The sequence of gates to apply.
        """
        for gate in gates:
            self.apply_gate(*gate)

        self._psi.squeeze_()

    def apply_circuit(self, gates):
        import warnings
        msg = ("``apply_circuit`` is deprecated in favour of ``apply_gates``.")
        warnings.warn(msg, DeprecationWarning)
        self.apply_gates(gates)

    def h(self, i, gate_round=None):
        self.apply_gate('H', i, gate_round=gate_round)

    def x(self, i, gate_round=None):
        self.apply_gate('X', i, gate_round=gate_round)

    def y(self, i, gate_round=None):
        self.apply_gate('Y', i, gate_round=gate_round)

    def z(self, i, gate_round=None):
        self.apply_gate('Z', i, gate_round=gate_round)

    def s(self, i, gate_round=None):
        self.apply_gate('S', i, gate_round=gate_round)

    def t(self, i, gate_round=None):
        self.apply_gate('T', i, gate_round=gate_round)

    def x_1_2(self, i, gate_round=None):
        self.apply_gate('X_1_2', i, gate_round=gate_round)

    def y_1_2(self, i, gate_round=None):
        self.apply_gate('Y_1_2', i, gate_round=gate_round)

    def z_1_2(self, i, gate_round=None):
        self.apply_gate('Z_1_2', i, gate_round=gate_round)

    def w_1_2(self, i, gate_round=None):
        self.apply_gate('W_1_2', i, gate_round=gate_round)

    def hz_1_2(self, i, gate_round=None):
        self.apply_gate('HZ_1_2', i, gate_round=gate_round)

    # constant two qubit gates

    def cnot(self, i, j, gate_round=None):
        self.apply_gate('CNOT', i, j, gate_round=gate_round)

    def cx(self, i, j, gate_round=None):
        self.apply_gate('CX', i, j, gate_round=gate_round)

    def cy(self, i, j, gate_round=None):
        self.apply_gate('CY', i, j, gate_round=gate_round)

    def cz(self, i, j, gate_round=None):
        self.apply_gate('CZ', i, j, gate_round=gate_round)

    def iswap(self, i, j, gate_round=None):
        self.apply_gate('ISWAP', i, j)

    # special non-tensor gates

    def iden(self, i, gate_round=None):
        pass

    def swap(self, i, j, gate_round=None):
        self.apply_gate('SWAP', i, j)

    # parametrizable gates

    def rx(self, theta, i, gate_round=None, parametrize=False):
        self.apply_gate('RX', theta, i, gate_round=gate_round,
                        parametrize=parametrize)

    def ry(self, theta, i, gate_round=None, parametrize=False):
        self.apply_gate('RY', theta, i, gate_round=gate_round,
                        parametrize=parametrize)

    def rz(self, theta, i, gate_round=None, parametrize=False):
        self.apply_gate('RZ', theta, i, gate_round=gate_round,
                        parametrize=parametrize)

    def u3(self, theta, phi, lamda, i, gate_round=None, parametrize=False):
        self.apply_gate('U3', theta, phi, lamda, i,
                        gate_round=gate_round, parametrize=parametrize)

    def fsim(self, theta, phi, i, j, gate_round=None, parametrize=False):
        self.apply_gate('FSIM', theta, phi, i, j,
                        gate_round=gate_round, parametrize=parametrize)

    @property
    def psi(self):
        """Tensor network representation of the wavefunction.
        """
        # make sure all same dtype and drop singlet dimensions
        psi = self._psi.squeeze()
        psi.astype_(psi.dtype)
        return psi

    @property
    def uni(self):
        """Tensor network representation of the unitary operator (i.e. with
        the initial state removed).
        """
        U = self.psi

        # rename the initial state rand_uuid bonds to 1D site inds
        ixmap = {f'k{i}': f'b{i}' for i in range(self.N)}

        # the first `N` tensors should be the tensors of input state
        tids = tuple(U.tensor_map)[:self.N]
        for i, tid in enumerate(tids):
            t = U._pop_tensor(tid)
            assert U.site_tag(i) in t.tags
            old_ix, = t.inds
            ixmap[old_ix] = f'k{i}'

        return U.reindex_(ixmap)

    def to_dense(self, reverse=False, dtype=None,
                 rank_simplify=False, **contract_opts):
        """Generate the dense representation of the final wavefunction.

        Parameters
        ----------
        reverse : bool, optional
            Whether to reverse the order of the subsystems, to match the
            convention of qiskit for example.
        contract_opts
            Suppled to :func:`~quimb.tensor.tensor_core.tensor_contract`.
        dtype : dtype or str, optional
            If given, convert the tensors to this dtype prior to contraction.
        rank_simplify : bool, optional
            If the network is complex, performing rank simplification first can
            aid the contraction path finding.

        Returns
        -------
        psi : qarray
            The densely represented wavefunction with ``dtype`` data.
        """
        psi = self.psi

        if dtype is not None:
            psi.astype_(dtype)

        if rank_simplify:
            psi.rank_simplify_()

        inds = [psi.site_ind(i) for i in range(self.N)]

        if reverse:
            inds = inds[::-1]

        p_dense = psi.to_dense(inds, tags=all, **contract_opts)
        return p_dense

    def simulate_counts(self, C, seed=None, reverse=False, **contract_opts):
        """Simulate measuring each qubit in the computational basis. See
        :func:`~quimb.calc.simulate_counts`.

        .. warning::

            This currently constructs the full wavefunction in order to sample
            the probabilities accurately.

        Parameters
        ----------
        C : int
            The number of 'experimental runs', i.e. total counts.
        seed : int, optional
            A seed for reproducibility.
        reverse : bool, optional
            Whether to reverse the order of the subsystems, to match the
            convention of qiskit for example.
        contract_opts
            Suppled to :func:`~quimb.tensor.tensor_core.tensor_contract`.

        Returns
        -------
        results : dict[str, int]
            The number of recorded counts for each
        """
        p_dense = self.to_dense(reverse=reverse, **contract_opts)
        return qu.simulate_counts(p_dense, C=C, seed=seed)

    def schrodinger_contract(self, *args, **contract_opts):
        ntensor = self._psi.num_tensors
        path = [(0, 1)] + [(0, i) for i in reversed(range(1, ntensor - 1))]
        return self.psi.contract(*args, optimize=path, **contract_opts)

    def update_params_from(self, tn):
        """Assuming ``tn`` is a tensor network with tensors tagged ``GATE_{i}``
        corresponding to this circuit (e.g. from ``circ.psi`` or ``circ.uni``)
        but with updated parameters, update the current circuit parameters and
        tensors with those values.

        This is an inplace modification of the ``Circuit``.

        Parameters
        ----------
        tn : TensorNetwork
            The tensor network to find the updated parameters from.
        """
        for i, gate in enumerate(self.gates):
            label = gate[0]
            tag = f'GATE_{i}'
            t = tn[tag]

            # sanity check that tensor(s) `t` correspond to the correct gate
            if label not in get_tags(t):
                raise ValueError(f"The tensor(s) correponding to gate {i} "
                                 f"should be tagged with '{label}', got {t}.")

            # only update gates and tensors if they are parametrizable
            if isinstance(t, PTensor):

                # update the actual tensor
                self._psi[tag].params = t.params

                # update the gate entry
                if label in ONE_QUBIT_PARAM_GATES:
                    new_gate = (label, *t.params, gate[-1])
                elif label in TWO_QUBIT_PARAM_GATES:
                    new_gate = (label, *t.params, *gate[-2:])
                else:
                    raise ValueError(f"Didn't recognise '{label}' "
                                     "gate as parametrizable.")

                self.gates[i] = new_gate

    def __repr__(self):
        r = "<Circuit(n={}, n_gates={}, gate_opts={})>"
        return r.format(self.N, len(self.gates), self.gate_opts)


class CircuitMPS(Circuit):
    """Quantum circuit simulation keeping the state always in a MPS form. If
    you think the circuit will not build up much entanglement, or you just want
    to keep a rigorous handle on how much entanglement is present, this can
    be useful.
    """

    def __init__(self, N=None, psi0=None, gate_opts=None, tags=None):
        gate_opts = {} if gate_opts is None else dict(gate_opts)
        gate_opts.setdefault('contract', 'swap+split')
        super().__init__(N, psi0, gate_opts, tags)

    @property
    def psi(self):
        # no squeeze so that bond dims of 1 preserved
        return self._psi


class CircuitDense(Circuit):
    """Quantum circuit simulation keeping the state in full dense form.
    """

    def __init__(self, N=None, psi0=None, gate_opts=None, tags=None):
        gate_opts = {} if gate_opts is None else dict(gate_opts)
        gate_opts.setdefault('contract', True)
        super().__init__(N, psi0, gate_opts, tags)

    @property
    def psi(self):
        return self._psi ^ all



"""Manages the spawning of mpi processes to send to the various solvers.
"""

import os
import functools

import numpy as np

from .slepc_linalg import (
    eigs_slepc, svds_slepc, mfn_multiply_slepc, ssolve_slepc,
)
from ..core import _NUM_THREAD_WORKERS

# Work out if already running as mpi
if ('OMPI_COMM_WORLD_SIZE' in os.environ) or ('PMI_SIZE' in os.environ):
    ALREADY_RUNNING_AS_MPI = True
    if '_QUIMB_MPI_LAUNCHED' not in os.environ:
        raise RuntimeError(
            "For the moment, quimb programs launched explicitly"
            " using MPI need to use `quimb-mpi-python`.")
    USE_SYNCRO = "QUIMB_SYNCRO_MPI" in os.environ
else:
    ALREADY_RUNNING_AS_MPI = False
    USE_SYNCRO = False

# Work out the desired total number of workers
for _NUM_MPI_WORKERS_VAR in ['QUIMB_NUM_MPI_WORKERS',
                             'QUIMB_NUM_PROCS',
                             'OMPI_COMM_WORLD_SIZE',
                             'PMI_SIZE',
                             'OMP_NUM_THREADS']:
    if _NUM_MPI_WORKERS_VAR in os.environ:
        NUM_MPI_WORKERS = int(os.environ[_NUM_MPI_WORKERS_VAR])
        break
else:
    import psutil
    _NUM_MPI_WORKERS_VAR = 'psutil'
    NUM_MPI_WORKERS = psutil.cpu_count(logical=False)


def bcast(result, comm, result_rank):
    """Broadcast a result to all workers, dispatching to proper MPI (rather
    than pickled) communication if the result is a numpy array.
    """
    rank = comm.Get_rank()

    # make sure all workers know if result is an array or not
    if rank == result_rank:
        is_ndarray = isinstance(result, np.ndarray)
    else:
        is_ndarray = None
    is_ndarray = comm.bcast(is_ndarray, root=result_rank)

    # standard (pickle) bcast if not array
    if not is_ndarray:
        return comm.bcast(result, root=result_rank)

    # make sure all workers have shape and dtype
    if rank == result_rank:
        shape_dtype = result.shape, str(result.dtype)
    else:
        shape_dtype = None

    shape_dtype = comm.bcast(shape_dtype, root=result_rank)
    shape, dtype = shape_dtype

    # allocate data space
    if rank != result_rank:
        result = np.empty(shape, dtype=dtype)

    # use fast communication for main array
    comm.Bcast(result, root=result_rank)

    return result


class SyncroFuture:

    def __init__(self, result, result_rank, comm):
        self._result = result
        self.result_rank = result_rank
        self.comm = comm

    def result(self):
        rank = self.comm.Get_rank()

        if rank == self.result_rank:
            should_it = (isinstance(self._result, tuple) and
                         any(isinstance(x, np.ndarray) for x in self._result))
            if should_it:
                iterate_over = len(self._result)
            else:
                iterate_over = 0
        else:
            iterate_over = None
        iterate_over = self.comm.bcast(iterate_over, root=self.result_rank)

        if iterate_over:
            if rank != self.result_rank:
                self._result = (None,) * iterate_over

            result = tuple(bcast(x, self.comm, self.result_rank)
                           for x in self._result)
        else:
            result = bcast(self._result, self.comm, self.result_rank)

        return result

    @staticmethod
    def cancel():
        raise ValueError("SyncroFutures cannot be cancelled - they are "
                         "submitted in a parallel round-robin fasion where "
                         "each worker immediately computes all its results.")


class SynchroMPIPool:

    def __init__(self):
        import itertools
        from mpi4py import MPI
        self.comm = MPI.COMM_WORLD
        self.size = self.comm.Get_size()
        self.rank = self.comm.Get_rank()
        self.counter = itertools.cycle(range(0, NUM_MPI_WORKERS))
        self._max_workers = self.size

    def submit(self, fn, *args, **kwargs):
        # round robin iterate through ranks
        current_counter = next(self.counter)

        # accept job and compute if have the same rank, else do nothing
        if current_counter == self.rank:
            res = fn(*args, **kwargs)
        else:
            res = None

        # wrap the result in a SyncroFuture, that will broadcast result
        return SyncroFuture(res, current_counter, self.comm)

    def shutdown(self):
        pass


class CachedPoolWithShutdown:
    """Decorator for caching the mpi pool when called with the equivalent args,
    and shutting down previous ones when not needed.
    """

    def __init__(self, pool_fn):
        self._settings = '__UNINITIALIZED__'
        self._pool_fn = pool_fn

    def __call__(self, num_workers=None, num_threads=1):
        # convert None to default so the cache the same
        if num_workers is None:
            num_workers = NUM_MPI_WORKERS
        elif ALREADY_RUNNING_AS_MPI and (num_workers != NUM_MPI_WORKERS):
            raise ValueError("Can't specify number of processes when running "
                             "under MPI rather than spawning processes.")

        # first call
        if self._settings == '__UNINITIALIZED__':
            self._pool = self._pool_fn(num_workers, num_threads)
            self._settings = (num_workers, num_threads)
        # new type of pool requested
        elif self._settings != (num_workers, num_threads):
            self._pool.shutdown()
            self._pool = self._pool_fn(num_workers, num_threads)
            self._settings = (num_workers, num_threads)
        return self._pool


@CachedPoolWithShutdown
def get_mpi_pool(num_workers=None, num_threads=1):
    """Get the MPI executor pool, with specified number of processes and
    threads per process.
    """
    if (num_workers == 1) and (num_threads == _NUM_THREAD_WORKERS):
        from concurrent.futures import ProcessPoolExecutor
        return ProcessPoolExecutor(1)

    if USE_SYNCRO:
        return SynchroMPIPool()

    from mpi4py.futures import MPIPoolExecutor
    return MPIPoolExecutor(num_workers, main=False,
                           env={'OMP_NUM_THREADS': str(num_threads),
                                'QUIMB_NUM_MPI_WORKERS': str(num_workers),
                                '_QUIMB_MPI_LAUNCHED': 'SPAWNED'})


class GetMPIBeforeCall(object):
    """Wrap a function to automatically get the correct communicator before
    its called, and to set the `comm_self` kwarg to allow forced self mode.

    This is called by every mpi process before the function evaluation.
    """

    def __init__(self, fn):
        self.fn = fn

    def __call__(self, *args,
                 comm_self=False,
                 wait_for_workers=None,
                 **kwargs):
        """
        Parameters
        ----------
        *args :
            Supplied to self.fn
        comm_self : bool, optional
            Whether to force use of MPI.COMM_SELF
        wait_for_workers : int, optional
            If set, wait for the communicator to have this many workers, this
            can help to catch some errors regarding expected worker numbers.
        **kwargs :
            Supplied to self.fn
        """
        from mpi4py import MPI

        if not comm_self:
            comm = MPI.COMM_WORLD
        else:
            comm = MPI.COMM_SELF

        if wait_for_workers is not None:
            from time import time
            t0 = time()
            while comm.Get_size() != wait_for_workers:
                if time() - t0 > 2:
                    raise RuntimeError(
                        f"Timeout while waiting for {wait_for_workers} "
                        f"workers to join comm {comm}.")

        comm.Barrier()
        res = self.fn(*args, comm=comm, **kwargs)
        comm.Barrier()
        return res


class SpawnMPIProcessesFunc(object):
    """Automatically wrap a function to be executed in parallel by a
    pool of mpi workers.

    This is only called by the master mpi process in manual mode, only by
    the (non-mpi) spawning process in automatic mode, or by all processes in
    syncro mode.
    """

    def __init__(self, fn):
        self.fn = fn

    def __call__(self, *args,
                 num_workers=None,
                 num_threads=1,
                 mpi_pool=None,
                 spawn_all=USE_SYNCRO or (not ALREADY_RUNNING_AS_MPI),
                 **kwargs):
        """
        Parameters
        ----------
            *args
                Supplied to `self.fn`.
            num_workers : int, optional
                How many total process should run function in parallel.
            num_threads : int, optional
                How many (OMP) threads each process should use
            mpi_pool : pool-like, optional
                If not None (default), submit function to this pool.
            spawn_all : bool, optional
                Whether all the parallel processes should be spawned (True), or
                num_workers - 1, so that the current process can also do work.
            **kwargs
                Supplied to `self.fn`.

        Returns
        -------
            `fn` output from the master process.
        """
        if num_workers is None:
            num_workers = NUM_MPI_WORKERS

        if num_workers == 1:  # no pool or communicator required
            return self.fn(*args, comm_self=True, **kwargs)

        kwargs['wait_for_workers'] = num_workers

        if mpi_pool is not None:
            pool = mpi_pool
        else:
            pool = get_mpi_pool(num_workers, num_threads)

        # the (non mpi) main process is idle while the workers compute.
        if spawn_all:
            futures = [pool.submit(self.fn, *args, **kwargs)
                       for _ in range(num_workers)]
            results = [f.result() for f in futures]

        # the master process is the master mpi process and contributes
        else:
            futures = [pool.submit(self.fn, *args, **kwargs)
                       for _ in range(num_workers - 1)]
            results = ([self.fn(*args, **kwargs)] +
                       [f.result() for f in futures])

        # Get master result, (not always first submitted)
        return next(r for r in results if r is not None)


# ---------------------------------- SLEPC ---------------------------------- #

eigs_slepc_mpi = functools.wraps(eigs_slepc)(
    GetMPIBeforeCall(eigs_slepc))
eigs_slepc_spawn = functools.wraps(eigs_slepc)(
    SpawnMPIProcessesFunc(eigs_slepc_mpi))

svds_slepc_mpi = functools.wraps(svds_slepc)(
    GetMPIBeforeCall(svds_slepc))
svds_slepc_spawn = functools.wraps(svds_slepc)(
    SpawnMPIProcessesFunc(svds_slepc_mpi))

mfn_multiply_slepc_mpi = functools.wraps(mfn_multiply_slepc)(
    GetMPIBeforeCall(mfn_multiply_slepc))
mfn_multiply_slepc_spawn = functools.wraps(mfn_multiply_slepc)(
    SpawnMPIProcessesFunc(mfn_multiply_slepc_mpi))

ssolve_slepc_mpi = functools.wraps(ssolve_slepc)(
    GetMPIBeforeCall(ssolve_slepc))
ssolve_slepc_spawn = functools.wraps(ssolve_slepc)(
    SpawnMPIProcessesFunc(ssolve_slepc_mpi))
